package roboforge.ui;

/*
* File:
* BApplication.java
*
* Description:
* The bootstrap application class. Performs initialization
* functions, then loads the main screen.
*
* Modification History: ( Name, Date, Notes )
* Created SJS, 27/01/2000.
*
*/

import javax.swing.*;

import roboforge.ui.fusion.*;
import roboforge.ui.components.BDialog;
import roboforge.ui.components.BPromptDialog;
//import roboforge.ui.events.*;
import roboforge.util.processors.*;

import roboforge.BDebug;
import roboforge.BGlobal;
import roboforge.BTextures;
import roboforge.sound.*;

//import java.awt.GraphicsEnvironment;
import java.awt.Cursor;

//import javax.vecmath.*;
//import javax.media.j3d.*;


public class BApplication
{

  // Class variables.
  // -------------------------------------------------------------

  // A single instance of BApplication exists at any given time.
  protected static BApplication instance = null;

  // Log file name...
  protected static final String LOGFILE = "output.txt";

  
  
  // Main method...
  // -------------------------------------------------------------
  
  public static void main(String[] args)
  {
    //BDebug.print("user.dir:  "+ System.getProperty("user.dir"));
    //BDebug.print("user.home: "+ System.getProperty("user.home"));

    // Load up BGlobals from file..
    BGlobal.init();

    // Set up log file...
    try
      { BSystemLogger.start(BGlobal.USERPATH+ LOGFILE); }
    catch (Exception e)
    {
      //e.printStackTrace();
      abort();
    }

    // Do not perform deprecated behavior (old interface stuff.)
    BGlobal.DEPRECATED = false;

    //VirtualUniverse.setJ3DThreadPriority(Thread.MIN_PRIORITY );

    instance = new BApplication(args);
  }


  // Internal variables.
  // -------------------------------------------------------------

  protected BWindow win;

  public static BWindow getWindow()
    { return instance.win; }


  // Constructor.
  // -------------------------------------------------------------

  public BApplication(String[] args)
  {
    // Parse command-line arguments to the game...
    parseArguments(args);

    // Open up the main screen...
    SwingUtilities.invokeLater(new BMainScreenLoader());
  }


  protected class BMainScreenLoader implements Runnable
  {
    public void run()
    {
      // Start up the sound system...
      BSoundSystem.start();

      // Install our custom look & feel.
      initializeSwing();

      // Create the application's wrapper window.
      win = new BWindow();

      // Initialize Java 3D rendering...
      initialize3DRendering();
      
      // Initialize main screen.
      //
      BMainScreen mainScreen = new BMainScreen();

      // When loaded, display the main screen.
      win.push(mainScreen);
      mainScreen = null;

      // Display the window.
      win.setVisible(true);
    }
  }


  // Command-line arguments.
  // -------------------------------------------------------------

  protected void parseArguments(String[] args)
    // Parses arguments supplied to the game
    // from the command-line.
  {
    for (int i=0; i<args.length; i++)
      parseArgument(args[i]);
  }

  protected void parseArgument(String arg)
  {
    // if (arg.equals("devmode"))
    //   BGlobal.DEVELOPER_MODE = true;
  }


  // Shutdown methods.
  // -------------------------------------------------------------

  public static void restart()
    { shutdown(-2); }
  
  public static void shutdown()
    { shutdown(0); }

  public static void shutdown(int code)
  {
    BDebug.print("BApplication.onShutdown");

    // Tell window to close (along with any open screens.)
    instance.win.onClose();

    // Perform application clean-up.
    onClose();

    // Exit the program.
    System.exit(code);
  }

  public static void abort()
    { abort(null); }

  public static void abort(String msg)
  {
    try
    {
      String abortMessage;

      if (msg != null)
        abortMessage = "Fatal error: "+ msg;
      else
        abortMessage = "A fatal error has occurred. Please see "+
          BGlobal.USERPATH+ LOGFILE+ " for details.";

      System.out.println(abortMessage);


      if (instance.win != null)
        JOptionPane.showMessageDialog
        (
          instance.win,
          abortMessage,
          "Fatal Error",
          JOptionPane.PLAIN_MESSAGE
        );
    }
    catch (Exception e)
      { }
    finally
    {

      // Close down vital resources...
      onClose();

      // Exit the program.
      System.exit(1);
    }
  }


  // Activation methods.
  // -------------------------------------------------------------

  public static void onClose()
  {
    BDebug.print("BApplication.onClose");

    // Shut down the sound system...
    BSoundSystem.stop();

    // Shut down the system log-file...
    BSystemLogger.stop();
  }


  // Initialization helper methods...
  // -------------------------------------------------------------

  protected void initialize3DRendering()
    { BTextures.setListener(win); }

  protected void initializeSwing()
  {
    BDebug.seedTimer();

    // Make tooltips heavyweight.
    // TODO: What's up with this deprecation crap?
    ToolTipManager tooltips = ToolTipManager.sharedInstance();

    tooltips.setLightWeightPopupEnabled(false);
    tooltips.setInitialDelay(500);
    tooltips.setReshowDelay(500);

    // Make popup menus heavyweight.
    JPopupMenu.setDefaultLightWeightPopupEnabled(false);

    // Disable the swing focus manager, since
    // we mix heavy and lightweight components.
    // FocusManager.disableSwingFocusManager();


    boolean success =
      FusionLookAndFeel.install();

    if (!success)
      abort("Could not install L&F.");

    BDebug.printTime("initializeSwing()");
  }


  // Windowing methods...
  // -------------------------------------------------------------

  protected static BPromptDialog notifyDialog;


  public static void openScreen(BScreen screen)
    { instance.win.push(screen); }

  public static boolean closeScreen(BScreen screen)
    { return instance.win.remove(screen, true); }
  
  public static boolean closeScreen()
    { return closeScreen(instance.win.peek()); }
 
  public static boolean hideScreen(BScreen screen)
    { return instance.win.remove(screen, false); }
  
  public static boolean hideScreen()
    { return hideScreen(instance.win.peek()); }

  
  public static void setCursor(Cursor c)
    { instance.win.setCursor(c); }

  public static void setBusy(boolean isBusy)
  {
    if (isBusy)
      instance.win.setBusyCursor();
    else
      instance.win.setCursor();

    instance.win.setInputEnabled(!isBusy);
  }

  public static void notifyUser(String msg)
    { notifyUser("", msg); }

  public static void notifyUser(String title, String msg)
  {
    if (notifyDialog == null)
      notifyDialog = BPromptDialog.createNotifyDialog();

    notifyDialog.setTitle(title);
    notifyDialog.setMessage(msg);

    BProcessor p = new BProcessor()
    {
      public void process()
        { instance.win.showDialog(notifyDialog); }
    };

    p.start();
  }


  public static void showDialog(BDialog dialog)
    { instance.win.showDialog(dialog); }

  public static void hideDialog(BDialog d)
    { instance.win.hideDialog(d); }

  public static void setInputEnabled(boolean enabled)
    { instance.win.setInputEnabled(enabled); }


}
