package roboforge.ui;

import roboforge.BArena;
import roboforge.BArenaObjectPosition;
import roboforge.BGlobal;
import roboforge.ai.orient.*;
import roboforge.ui.components.*;
import roboforge.ui.modes.*;
import roboforge.math.BMath;
import roboforge.BBot;
import roboforge.BDebug;

import java.awt.*;
import java.awt.event.*;

import javax.swing.Icon;
import javax.vecmath.*;

////////////////////////////////////////////////////////////////////////////////
//
// class BArenaSetupPicker provides a tool for setting up a simulation
// situation before testing
//
//
//
public class BArenaSetupPicker extends BMouseableGrid2D
{

  // Class variables...
  // -------------------------------------------------------------
  protected static final double MIN_ORIENT = -30;
  protected static final double MAX_ORIENT =  30;

  protected static RenderingHints hints = new RenderingHints (
    RenderingHints.KEY_ANTIALIASING,
    RenderingHints.VALUE_ANTIALIAS_ON );


  // Member variables...
  // -------------------------------------------------------------
  protected BModeMulti             viewMode;
  protected BModeSetup             setupMode;

  protected Color            lineColor = BGlobal.colors.getDarkShade();

  protected BBotSymbol heroSymbol = new BBotSymbol(true);
  protected BBotSymbol enemySymbol = new BBotSymbol(true);

  protected Point2d enemy   = new Point2d(0,0);
  protected Point2d hero    = new Point2d(0,0);
  protected Point2d origin  = new Point2d(0,0);
  protected float   facing  = 0;
  protected float   angle   = 0;

  protected float   heroFacing = 0.0f;
  protected float   enemyFacing = 0.0f;

  protected Point2d o       = new Point2d(0,0);
  protected Point2d e       = new Point2d(0,0);
  protected Point2d p       = new Point2d(0,0);

  protected Vector2d delta  = new Vector2d(0,0);
  protected Vector2d up     = new Vector2d(0,-1);

  protected BArena arena;
  protected BBot botHero;
  protected BBot botEnemy;

  // Constructors...
  // -------------------------------------------------------------
  public BArenaSetupPicker(BUIResources res, String key)
  {
    super(res, key);

    viewMode = viewModeWrapper(new BModeMouse());
    setMode(viewMode);

    setupMode     = new BModeSetup();

    this.addMouseListener(setupMode);
    this.addMouseMotionListener(setupMode);

  }

  // set the arena
  public void setArena(BArena arena)
  {
    this.arena = arena;
    repaint();
  }

  // set the bots
  public void setBots(BBot botHero, BBot botEnemy)
  {
    this.botHero = botHero;
    this.botEnemy = botEnemy;

    setBotOrient();

  }

  // set the symbol orientations to match the bots
  public void setBotOrient()
  {
    if(botHero != null && botEnemy != null)
    {
      hero.x = BGlobal.CANVAS_TO_WORLD_SCALE * botHero.getArenaPosition().getX();
      hero.y = BGlobal.CANVAS_TO_WORLD_SCALE * botHero.getArenaPosition().getZ();
      enemy.x = BGlobal.CANVAS_TO_WORLD_SCALE * botEnemy.getArenaPosition().getX();
      enemy.y = BGlobal.CANVAS_TO_WORLD_SCALE * botEnemy.getArenaPosition().getZ();

      heroFacing = (float)Math.toDegrees(botHero.getArenaPosition().getFacing());
      enemyFacing = (float)Math.toDegrees(botEnemy.getArenaPosition().getFacing());

      setupMode.calculateAngle();
      repaint();
    }
  }


  public void setViewLocation(Point2d p)
  {
    p.x = constrain(p.x);
    p.y = constrain(p.y);

    super.setViewLocation(p);
  }

  protected double constrain(double n)
    { return Math.min(Math.max(n, MIN_ORIENT), MAX_ORIENT); }

  // Paint method.
  // -------------------------------------------------------------

  public void paint(Graphics g)
  {
    super.paint(g);

    Graphics2D g2D =
      (Graphics2D) g;

    if (setupMode != null)
      setupMode.paint(g2D);
  }


  // takes care of setting up the arena - enables the setting of
  // position and rotation of bots
  protected class BModeSetup extends BModeGrid implements MouseListener, MouseMotionListener
  {

   protected Point2d mouse = new Point2d();

   protected boolean heroSelected = false;
   protected boolean enemySelected = false;

   public void paint(Graphics2D g)
    {
      g.setRenderingHints(hints);

      metersToPels(origin, o);
      metersToPels(enemy, e);
      metersToPels(hero, p);

      int ox = (int) o.x;
      int oy = (int) o.y;
      int px = (int) p.x;
      int py = (int) p.y;
      int ex = (int) e.x;
      int ey = (int) e.y;

      int eStrx = ox + (int) metersToXPels(2);
      g.setColor(lineColor);

      paintArenaEdge(g, ox, oy);

      // Draw bots...
      paintBot(g, enemySymbol, botEnemy, ex, ey, (float)Math.toRadians(enemyFacing));
      paintBot(g, heroSymbol, botHero, px, py, (float)Math.toRadians(heroFacing));

      Dimension d = getSize();
    }

    // draws a bot
    protected void paintBot(Graphics2D g, BBotSymbol bot, BBot realBot, int x, int y, float theta)
    {
      double sx = pelsPerXMeter / 12.5;
      double sy = pelsPerYMeter / 12.5;

      g.translate(x, y);
      g.scale(sx, sy);
      g.rotate(-theta - Math.PI);

      boolean flip = false;



      if(realBot == botHero)
      {
        bot.paint(g, theta, realBot.getName(), "Distance: " + BMath.format(BMath.round(hero.distance(enemy))) + "m", "Angle: " + BMath.format(BMath.round(angle)) + ".0" , true);
      }
      else
      {
        bot.paint(g, theta, "", realBot.getName(), "", true);
      }

      g.rotate(theta + Math.PI);
      g.scale(1/sx, 1/sy);
      g.translate(-x, -y);
    }

    // draws the circle that represents the edge of the arena
    protected void paintArenaEdge(Graphics2D g, int x, int y)
    {
      int arenaSize = 0;

      if(arena != null)
        arenaSize = arena.getSize();

      int width = (int)metersToXPels(arenaSize);
      int height = (int)metersToYPels(arenaSize);
      g.drawOval((int)(x - width / 2), (int)(y - height / 2), width, height);
    }

    protected float setFacing(float f)
    {
      float fa = f;
      if (fa < 0)
        fa = (fa % 360) + 360;
      else if (fa > 360)
        fa = fa % 360;

      return fa;
    }

    protected void calculateAngle()
    {
      delta.sub(hero, enemy);
      angle = (float) delta.angle(up);

      if (hero.x > enemy.x)
        angle = -angle;

      angle -= Math.toRadians(heroFacing);
      angle = (float)Math.toDegrees(angle);
      angle = setFacing(angle);
      angle = 360 - angle;
    }

    // check to see if a point is within the arena
    protected boolean withinArena(double x, double y)
    {
      double distance = Math.sqrt( Math.pow( x, 2) + Math.pow( y, 2));
      if (distance * 2 > arena.getSize())
        return false;
      else
        return true;
    }

    // mouse listeners
    public void mousePressed(MouseEvent e)
    {

      super.mousePressed(e);
      mouse.set(x, y);
      pelsToMeters(mouse, mouse);

      double distanceHero = Math.sqrt( Math.pow( (mouse.x - hero.x), 2) + Math.pow( (mouse.y - hero.y), 2));
      double distanceEnemy = Math.sqrt( Math.pow( (mouse.x - enemy.x), 2) + Math.pow( (mouse.y - enemy.y), 2));

      if (distanceHero <= 2)
      {
        heroSymbol.setCol(BGlobal.colors.getWhiteShade());
        enemySymbol.setCol(BGlobal.colors.getMidShade());
        heroSelected = true;
        enemySelected = false;
      }
      else if (distanceEnemy <= 2)
      {
        enemySymbol.setCol(BGlobal.colors.getWhiteShade());
        heroSymbol.setCol(BGlobal.colors.getMidShade());
        heroSelected = false;
        enemySelected = true;
      }
      else
      {
        enemySymbol.setCol(BGlobal.colors.getMidShade());
        heroSymbol.setCol(BGlobal.colors.getMidShade());
        heroSelected = false;
        enemySelected = false;
      }
      repaint();
    }

    public void mouseDragged(MouseEvent e)
    {
      super.mouseDragged(e);
      mouse.set(x, y);
      pelsToMeters(mouse, mouse);

      boolean isFacing = e.isMetaDown();

      if (isFacing)
      {
        if(heroSelected)
        {
          heroFacing = setFacing(heroFacing + dy);
        }
        else if(enemySelected)
        {
          enemyFacing = setFacing(enemyFacing + dy);
        }
      }
      else
      {
        if(heroSelected)
        {
          if (withinArena(mouse.x,  mouse.y))
          {
            hero.x = mouse.x;
            hero.y = mouse.y;
          }
        }
        else if(enemySelected)
        {
          if (withinArena(mouse.x,  mouse.y))
          {
            enemy.x = mouse.x;
            enemy.y = mouse.y;
          }
        }
      }
      calculateAngle();
      repaint();
    }

    public void mouseMoved(MouseEvent e) {}
    public void mouseEntered(MouseEvent e){}
    public void mouseExited(MouseEvent e){}
    public void mouseClicked(MouseEvent e){}

    public void mouseReleased(MouseEvent e)
    {
      super.mouseReleased(e);
      mouse.set(x, y);
      pelsToMeters(mouse, mouse);
      BArenaSetupPicker.this.setCursor(Cursor.getDefaultCursor());

      botHero.setArenaPosition(new BArenaObjectPosition((float)(BGlobal.WORLD_TO_CANVAS_SCALE * hero.x), 0.0f, (float)(BGlobal.WORLD_TO_CANVAS_SCALE * hero.y), (float)Math.toRadians(heroFacing)));
      botEnemy.setArenaPosition(new BArenaObjectPosition((float)(BGlobal.WORLD_TO_CANVAS_SCALE * enemy.x), 0.0f, (float)(BGlobal.WORLD_TO_CANVAS_SCALE * enemy.y), (float)Math.toRadians(enemyFacing)));

    }
  }
}
//
//
////////////////////////////////////////////////////////////////////////////////
