package roboforge.ui;

import roboforge.BDebug;
import roboforge.BMove;
import roboforge.BBot;

import java.util.Enumeration;
import javax.swing.event.*;
import javax.media.j3d.*;
import roboforge.util.timing.*;

public class BMoveBehavior extends Behavior
{

  // Constants.
  // -------------------------------------------------------------

  protected static final float SECOND               = 1000.0f;
  protected static final long  PLAY_TIMEOUT         = 50 * 1000;
  protected static final long  DEFAULT_DELTA_TIME   = 10;

  // Internal variables.
  // -------------------------------------------------------------

  protected B3DBotCanvas canvas                     = null;
  protected BBot bot                                = null;
  protected boolean isSimulating                    = false;
  protected boolean isPlaying                       = false;
  protected boolean isMoving                        = false;

  protected boolean isStopRequested                 = false;

  protected long clockTime, lastTime, elapsedTime, dT;

  protected BMove curMove                           = null;
  protected int curMoveIndex                        = 0;
  protected int newMoveIndex                        = 0;

  // Wakeup condition (wake up on every frame rendered.)
  protected WakeupOnElapsedFrames wFrame = new WakeupOnElapsedFrames(0);

  // Supports a single change listener.
  protected ChangeListener listener = null;


  // Constructor.
  // -------------------------------------------------------------

  public BMoveBehavior(B3DBotCanvas canvas, BBot bot)
  {
    this.canvas    = canvas;
    this.bot = bot;
  }

  // Accessor methods.
  // -------------------------------------------------------------

  public boolean isSimulating()
    { return isSimulating; }

  public boolean isPlaying()
    { return isPlaying; }

  public int getSnapshotIndex()
    { return curMoveIndex; }


  // Player commands.
  // -------------------------------------------------------------

  public void start(BMove move)
  {
    // BDebug.print("BMoveBehavior started; index = "+ move.getCurrentSnapshotIndex());

    this.curMove = move;
    curMoveIndex = move.getCurrentSnapshotIndex();

    bot.clearTippingFlag();
            
    lastTime    = (long)( BClock.time() * 1000);
    elapsedTime = 0;

    isSimulating = true;

    fireStateChanged();

    resume();
  }

  protected void resume()
  {
    if (isSimulating)
    {
      // BDebug.print("BMoveBehavior resumed.");
      this.setEnable(true);
      this.isPlaying = true;
    }
  }

  protected void pause()
  {
    if (isSimulating)
    {
      // BDebug.print("BMoveBehavior paused.");
      this.setEnable(false);
      this.isPlaying = false;
    }
  }

  public void stop()
  {
    // BDebug.print ("BMoveBehavior stop() signalled.");

    if (isSimulating && isPlaying)
      // Request a call to stopFromBehaviourThread().
      isStopRequested = true;
    else
      executeStopRequest();
  }

  protected void executeStopRequest()
  {
    // BDebug.print ("BMoveBehavior executeStopRequest().");

    this.isStopRequested = false;
    this.isSimulating    = false;
    this.isPlaying       = false;
    this.setEnable(false);

    fireStateChanged();
  }

  // Behavior methods.
  // -------------------------------------------------------------

  public void initialize()
  {
    // BDebug.print("BMoveBehavior.initialize.");
    this.setEnable(false);
    this.wakeupOn(wFrame);
  }

  public void processStimulus(Enumeration criteria)
  {
    // Get the current time...
    clockTime = (long)(BClock.time() * 1000);

    // Stop if requested to do so...
    if (isStopRequested)
    {
      executeStopRequest();
      wakeupOn(wFrame);
      return;
    }

    else if (isSimulating)
    {
      // Calculate size of simulator time-step...
      dT = clockTime - lastTime;

      if (dT != 0)
      {
        // Remember current clock-time for next frame.
        lastTime = clockTime;

        // Perform a simulation step.
        // BDebug.print("time slice: "+ dT+ "ms");
        //isMoving = simulator.startMoveSlice(dT / SECOND);
        isMoving = bot.executeSingleMove(curMove, dT / SECOND);


        // Check to see if move has completed or timed out...
        if ((!isMoving) || (elapsedTime >= PLAY_TIMEOUT))
          executeStopRequest();

        // Update canvas view of bot.
        canvas.setViewFromParameters();

        // Increment time elapsed.
        // TODO: This measure probably suffers from accumulated errors.
        elapsedTime += dT;

        // Check if we need to fire off a change-event...
        newMoveIndex = curMove.getCurrentSnapshotIndex();
        if (newMoveIndex != curMoveIndex)
        {
          // BDebug.print("BMoveBehavior: snapshot change ["+
          //   curMoveIndex+ ","+ newMoveIndex+ "].");

          curMoveIndex = newMoveIndex;
          fireStateChanged();
        }
      }
    }


    Thread.currentThread().yield();

    // Reset wakeup conditions.
    wakeupOn(wFrame);
  }

  public void addChangeListener(ChangeListener c)
    { listener = c; }

  protected void fireStateChanged()
  {
    if (listener != null)
      listener.stateChanged(new ChangeEvent(this));
  }

}


