package roboforge.ui;

import roboforge.BBot;
import roboforge.BGlobal;
import roboforge.BDebug;
import roboforge.BComponent;
import roboforge.BMove;
import roboforge.BMoveList;
import roboforge.BSnapshot;

import roboforge.math.*;
import roboforge.ui.action.*;
import roboforge.ui.components.*;
import roboforge.ui.components.border.*;
import roboforge.ui.components.text.*;
import roboforge.ui.events.*;
import roboforge.util.graphics.*;
import roboforge.util.text.*;

import java.awt.*;
import java.awt.event.*;
import java.util.Vector;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;

import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;
import javax.swing.table.*;



public class BMovesLibrary extends BToolPanel
  implements ListDataListener
{

  // Internal variables ...
  // -------------------------------------------------------------

  protected BBot bot                           = null;
  protected BMoveList moves                    = null;
  protected BMoveController control            = null;
  protected ListSelectionModel selectionModel;

  protected boolean wasActiveBeforeSave;


  // UI Resources ...
  // -------------------------------------------------------------

  protected BUIResources res = new BUIResources( new Object[]
  {
    "addMove.icon",               BIconCache.getHSB("add_icon_small.gif"),
    "addMove.tooltip",            "Add a new move.",

    "delMove.icon",               BIconCache.getHSB("del_icon_small.gif"),
    "delMove.tooltip",            "Remove this move.",

    "snapshots.autoscrolls",      Boolean.TRUE,
    "scroller.border",            new BLineBorder(colors.getDarkerShade()),
  });


  // Visual elements...
  // -------------------------------------------------------------

  protected BComboBox movesCombo            = new BComboBox();
  protected BNamedEditor movesEditor;
  protected BButton addMove                 = new BButton(res, "addMove");
  protected BButton delMove                 = new BButton(res, "delMove");
  protected AddMoveListener addMoveListener = new AddMoveListener();
  protected DelMoveListener delMoveListener = new DelMoveListener();

  protected BTable snapshots                = new BTable(res, "snapshots");
  protected BScrollPane scroller            = new BScrollPane(res, "scroller", snapshots);
  protected ControlDeck controlDeck;

  protected BSnapshotRenderer renderer;
  protected BSnapshotEditor editor;

  protected SelectionListener snapshotUpdateListener;

  // Constructor.
  // -------------------------------------------------------------

  public BMovesLibrary(B3DEditingPanel parent)
  {
    super(parent, "moves");

    this.bot         = parent.getBot();
    this.moves       = bot.getMoveList();
    this.control     = new BMoveController(bot, parent.getCanvas());
    this.controlDeck = new ControlDeck();

    moves.addListDataListener(this);

    layoutMovesLibraryChildren();
    configMovesLibraryChildren();

    res.cleanse();
    res = null;
  }

  // Children initialization methods...
  // -------------------------------------------------------------

  protected void layoutMovesLibraryChildren()
  {
    BBoxPanel box = createHBoxPanel();

    box.add(new BLabel("Move:"));
    box.addSmallMargin();
    box.addHGlued(movesCombo);
    box.addSmallMargin();
    box.add(addMove);
    box.addSmallMargin();
    box.add(delMove);

    add(box);
    addSmallMargin();

    add(scroller);
    addSmallMargin();

    add(controlDeck);
  }

  protected void configMovesLibraryChildren()
  {
    // Set up editable move names...
    movesEditor = new BNamedEditor(moves);
    movesEditor.setGuard(new BFilenameGuard());
      
    movesCombo.setEditable(true);
    movesCombo.setEditor(movesEditor);
    movesCombo.setModel(moves);
    
    

    addMove.addActionListener(addMoveListener);
    delMove.addActionListener(delMoveListener);

    selectionModel = control.getSelectionModel();
    selectionModel.setSelectionMode( ListSelectionModel.SINGLE_SELECTION );
    snapshots.setModel(control);
    snapshots.setSelectionModel(selectionModel);
    snapshots.setTableHeader(null);
    snapshots.setRowHeight(74);    
    
    renderer = new BSnapshotRenderer();
    snapshots.setDefaultRenderer(Object.class, renderer);

    editor = new BSnapshotEditor();
    snapshots.setDefaultEditor(Object.class, editor);

    SelectionListener snapshotUpdateListener = new SelectionListener();

    selectionModel.addListSelectionListener(snapshotUpdateListener);
    control.addTableModelListener(editor);
    control.addPropertyChangeListener(new ControlListener());
    
    onMoveChanged();
  }


  public void onSaveBegin()
  {
    wasActiveBeforeSave = control.isActive();

    if (wasActiveBeforeSave)
      control.onDeactivate();
  }

  public void onSaveEnd()
  {
    if (wasActiveBeforeSave)
      control.onActivate();
  }


  // UI Listeners...
  // -------------------------------------------------------------

  protected class AddMoveListener implements ActionListener
  {
    public void actionPerformed(ActionEvent e)
      { addMove(); }
  }

  protected class DelMoveListener implements ActionListener
  {
    public void actionPerformed(ActionEvent e)
    {
      BMove m = moves.getSelectedMove();
      if (m != null) moves.remove(m);
    }
  }


  // Data change methods...
  // -------------------------------------------------------------

  public void contentsChanged(ListDataEvent e)
    { onMoveChanged(); }

  public void intervalAdded(ListDataEvent e)
    { onMoveChanged(); }

  public void intervalRemoved(ListDataEvent e)
    { onMoveChanged(); }

  protected void onMoveChanged()
  {
    //BDebug.print ("onMoveChanged called ");
    BMove m = moves.getSelectedMove();



    if (m != control.getMove())
    {
      stopEditing();
      control.setMove(m);
      // YLY - HACK to get the combobox to update when a snapshot editor is active
      movesCombo.setSelectedItem(m);

    }

    setEnabledStatus();
  }

  
  protected void setEnabledStatus()
  {
    boolean movesExist = !moves.isEmpty();

    movesCombo.setEnabled(movesExist);
    delMove.setEnabled(movesExist);
    controlDeck.setEnabled(movesExist);
  }
  
  protected void stopEditing()
  {
    if (editor != null)
      editor.stopEditing();
  }

  
  
  
  // Move Control Listener...
  // -------------------------------------------------------------

  protected class ControlListener implements PropertyChangeListener
  {
    public void propertyChange(PropertyChangeEvent e)
    { 
      controlDeck.setControlsEnabled(); 
        
      boolean playing = control.isPlaying();
        
      parent.setWorking(playing);
      parent.setStatus(playing ?
        "Playing move '"+ control.getMove()+ "'..." :
        "Move '"+ control.getMove()+ "' complete.");
      
      if (playing)
        stopEditing();
    }
  }
  
  protected class SelectionListener implements ListSelectionListener
  {
    public void valueChanged(ListSelectionEvent e)
    {
     // BDebug.print ("BMovesLibrary.selectionListener called ");
      controlDeck.setControlsEnabled();

      int row = selectionModel.getMaxSelectionIndex();
      //  selectionModel.getMinSelectionIndex();

      Rectangle cell =
        snapshots.getCellRect(row, 0, true);

      snapshots.scrollRectToVisible(cell);
      editor.refresh();
    }
  }
  

  
  // -------------------------------------------------------------
  // Activation methods...
  // -------------------------------------------------------------

  public void onActivate()
  {
    super.onActivate();
    control.requestActivate();
  }

  public void onDeactivate()
  {
    control.requestDeactivate();
    super.onDeactivate();
  }

  public void onClose()
  {
    control.onClose();

    bot             = null;
    moves           = null;
    movesCombo      = null;
    movesEditor     = null;

    addMove.removeActionListener(addMoveListener);
    delMove.removeActionListener(delMoveListener);
    addMove         = null;
    delMove         = null;
    addMoveListener = null;
    delMoveListener = null;
    control         = null;

   // BDebug.print("BMovesLibrary.onClose() called ");
    snapshots.setSelectionModel(new DefaultListSelectionModel());
    selectionModel.removeListSelectionListener(snapshotUpdateListener);
    snapshotUpdateListener = null;
    selectionModel = null;

    editor.refresh.setAction(null);
    editor.chassis.setModel(new DefaultComboBoxModel());
    
    snapshots.setDefaultRenderer(Object.class, null);
    snapshots.setDefaultEditor(Object.class, null);
    editor = null;
    renderer = null;
    snapshots = null;




    super.onClose();
  }


  // Add/Remove moves...
  // -------------------------------------------------------------

  public void addMove()
  {
    int n = moves.size() + 1;
    BMove m = bot.createNewMove("Move "+ n);

    moves.addElement(m);
    moves.setSelectedItem(m);

    control.setNewMove(m);
  }



  // -------------------------------------------------------------
  // ControlDeck inner Class...
  // -------------------------------------------------------------

  protected class ControlDeck extends BBoxPanel
  {

    protected BUIResources res = new BUIResources( new Object[]
    {
      "play.icon",                  BIconCache.getHSB("play_icon.gif"),
      "play.tooltip",               "Play the current move.",

      "stop.icon",                  BIconCache.getHSB("stop_icon.gif"),
      "stop.tooltip",               "Stop playing the current move.",

      "insert.icon",                BIconCache.getHSB("snapshot_icon.gif"),
      "insert.tooltip",             "Capture a new snapshot.",

      "remove.icon",                BIconCache.getHSB("remove_snap_icon.gif"),
      "remove.tooltip",             "Delete the selected snapshot.",
    });


    protected BToggleButton play    = new BToggleButton(res, "play");
    protected BButton stop          = new BButton(res, "stop");
    protected BButton insert        = new BButton(res, "insert");
    protected BButton remove        = new BButton(res, "remove");


    public ControlDeck()
    {
      super(BoxLayout.X_AXIS);

      layoutControlDeckChildren();
      configControlDeckChildren();
    }

    protected void layoutControlDeckChildren()
    {
      addGlue();
      add(play);
      addSmallMargin();
      add(stop);
      addSmallMargin();
      add(insert);
      addSmallMargin();
      add(remove);
    }

    protected void configControlDeckChildren()
    {
      play.addActionListener(new ActionListener()
      { public void actionPerformed(ActionEvent e)
          { control.play(); play.setSelected(true); }});

      stop.addActionListener(new ActionListener()
      { public void actionPerformed(ActionEvent e)
          { control.stop(); }});

      insert.addActionListener(new ActionListener()
      {
        public void actionPerformed(ActionEvent e)
          {
            // BDebug.print ("ControlDeck.insert fired");
            control.record();

            editor.startEditing(control.getSelectedSnapshot());

          //  selectionModel.setAnchorSelectionIndex(1);
          //  selectionModel.setLeadSelectionIndex(1) ;
          //  selectionModel.setSelectionInterval(1, 1);

            //snapshotUpdateListener.valueChanged(   new ListSelectionEvent(this, selectionModel.getMaxSelectionIndex(), selectionModel.getMaxSelectionIndex(), true)     );
          }

      });

      remove.addActionListener(new ActionListener()
      { public void actionPerformed(ActionEvent e)
        { control.remove(); }});

      stop.setEnabled(false);
    }

    public void setControlsEnabled()
    {
      boolean playing       = control.isPlaying();
      boolean move_selected = (control.getMove() != null);
      boolean editable      = !playing && move_selected;

      play.setEnabled(move_selected);
      play.setSelected(playing);
      stop.setEnabled(playing);

      insert.setEnabled(editable);
      remove.setEnabled(editable && control.canRemoveSnap());
    }

    public void setEnabled(boolean enabled)
      { setControlsEnabled(); }
  }

  
  
  // -------------------------------------------------------------
  // Snapshot Cell Renderer inner class...
  // -------------------------------------------------------------
  
  
  protected class BSnapshotRenderer extends BBoxPanel
    implements TableCellRenderer
  {

    protected BUIResources res = new BUIResources( new Object[]
    {
      "thumb.size",           new Dimension(64, 64),
      "chassis.height",       new Integer(18),
      "chassis.font",         new Font("Dialog", Font.BOLD, 11),
      "power.height",         new Integer(16),

      // "delayLabel.height",    new Integer(16),

    });

    
    protected Color textSelectColor;
    protected Color selectColor;
    protected Color unselectColor;

    protected JLabel thumb      = new BLabel(res, "thumb");
    protected JLabel chassis    = new BLabel(res, "chassis");
    // protected JLabel delayLabel      = new BLabel(res, "delayLabel");

    protected JLabel power      = new BLabel(res, "power");
    protected ImageIcon icon    = new ImageIcon();


    public BSnapshotRenderer()
    {
      super(BBoxPanel.X_AXIS);

      layoutRenderer();
      configRenderer();
    }

    protected void layoutRenderer()
    {
      setEmptyBorder(5);
      add(thumb);
      addSmallMargin();
      
      BBoxPanel box = new BBoxPanel(BBoxPanel.Y_AXIS);
      box.setOpaque(false);
      box.addHGlued(chassis);
      box.addHGlued(power);
      // box.addHGlued(delayLabel);
      box.addGlue();
      
      add(box);
    }

    protected void configRenderer()
    {
      textSelectColor = colors.getMidShade();
      selectColor     = colors.getDarkShade();
      unselectColor   = colors.getVeryDarkShade();
    }


    public Component getTableCellRendererComponent
    (
      JTable table,
      Object value,
      boolean isSelected,
      boolean hasFocus,
      int row, int column
    )
    {
      BSnapshot snap = (BSnapshot) value;

      if (isSelected)
      {
        setBackground(selectColor);
        chassis.setForeground(textSelectColor);
        power.setForeground(textSelectColor);
        // delayLabel.setForeground(textSelectColor);
      }
      else
      {
        setBackground(unselectColor);
        chassis.setForeground(selectColor);
        power.setForeground(selectColor);
        // delayLabel.setForeground(selectColor);
      }

      Image img = snap.getThumbnail();

      if (img != null)
      {
        icon.setImage(img);
        thumb.setIcon(icon);
      }
      else
        thumb.setIcon(null);
        

      String chassisText, powerText;
      
      if (row == 0)
        chassisText = "Default position";
      else if (snap.getChassisMove() != null)
        chassisText =
            BStrings.toSentenceCase(snap.getChassisMove().toString())+ " "
          + BMath.format(snap.getMetersToMove())+ " "
          + snap.getChassisMoveUnits();
      else
        chassisText = "Chassis move: <none>";

      chassis.setText(chassisText);


      powerText = snap.getPowerDescription()+ " power";
      power.setText(powerText);

      return this;
    }

  }
  


  // -------------------------------------------------------------
  // Snapshot Cell Editor inner class...
  // -------------------------------------------------------------
  
  protected class BSnapshotEditor extends AbstractCellEditor
    implements TableCellEditor, ActionListener, TableModelListener
  {

    protected BUIResources res = new BUIResources( new Object[]
    {
      "container.size",       new Dimension(310, 74),
      "shield.size",          new Dimension(310, 74),
      "editor.size",          new Dimension(310, 74),

      "screen.size",          new Dimension(64, 64),
      "thumb.size",           new Dimension(64, 64),
      "thumb.location",       new Point(0, -1),
      "refresh.size",         new Dimension(20, 20),
      "refresh.location",     new Point(41, 40),

      "chassis.width",        new Integer(105),
      "amount.width",         new Integer(40),

      "units.size",           new Dimension(25, 20),
      "units.foreground",     colors.getMidShade(),

      "power.width",          new Integer(105),

      "chassisLabel.width",       new Integer(45),
      "chassisLabel.text",        "Chassis:",
      "chassisLabel.foreground",  colors.getMidShade(),

      "powerLabel.width",         new Integer(45),
      "powerLabel.text",          "Power:",
      "powerLabel.foreground",    colors.getMidShade(),

      // "delay.width",          new Integer(105),
      // "delay.foreground",     BGlobal.colors.getMidShade(),
      // "delay.text",           "Bot controller guides snapshot",
      // "delay.tooltip",        "The Bot's controller tries to finish moving all the joints simultaneously."

    });

    protected BSnapshot snap;
    protected boolean updateEnabled;
    
    protected Color selectColor  = colors.getDarkShade();
    
    protected BPanel      container     = new BPanel(res, "container");
    protected ShieldPanel shield        = new ShieldPanel(res, "shield");
    protected BBoxPanel   editor        = new BBoxPanel(res, "editor", BBoxPanel.X_AXIS);
    protected BBoxPanel   controls      = new BBoxPanel(BBoxPanel.Y_AXIS);
    protected BBoxPanel   chassisMotion = new BBoxPanel(BBoxPanel.X_AXIS);
     
    protected BPanel screen      = new BPanel(res, "screen");
    protected JLabel thumb       = new BLabel(res, "thumb");
    protected ImageIcon icon     = new ImageIcon();
    protected JButton refresh    = new BActionButton(res, "refresh");
    
    protected BComboBox chassis  = new BComboBox(res, "chassis");
    protected BFloatField amount = new BFloatField(res, "amount");
    protected JLabel units       = new BLabel(res, "units");

    protected BLabel chassisLabel = new BLabel(res, "chassisLabel");
    protected BLabel powerLabel   = new BLabel(res, "powerLabel");

    protected BComboBox power    = new BComboBox(res, "power");
    // protected BCheckBox delay    = new BCheckBox(res, "delay");


    public BSnapshotEditor()
    {
      super();

      layoutEditor();
      configEditor();
    }

    protected void layoutEditor()
    {
      screen.add(thumb);
      screen.addOnTop(refresh);
     
      chassisMotion.setBackground(selectColor);
      chassisMotion.add(chassisLabel);
      chassisMotion.add(chassis);
      chassisMotion.addTinyMargin();
      chassisMotion.add(amount);
      chassisMotion.addTinyMargin();
      chassisMotion.add(units);
      chassisMotion.addGlue();
      
      controls.setBackground(selectColor);
      controls.add(chassisMotion);
      controls.addTinyMargin();
      controls.addLinear(powerLabel, power, getHGlue());
      // controls.addLinear(delay, getHGlue());
      controls.addGlue();
      
      // editor.addBorderOutside(new LineBorder(colors.getDarkShade()));
      editor.setBackground(selectColor);
      editor.setEmptyBorder(5);
      editor.add(screen);
      editor.addSmallMargin();
      editor.add(controls);

      container.add(editor);
      container.addOnTop(shield);
    }

    protected void configEditor()
    {
      refresh.setAction(new SnapRefresher());
      //delay.setName("Delay Fast Joints");
      //delay.setText("Delay Fast Joints");
      //delay.setForeground(Color.white);

      Vector chassisMoves =
        bot.getChassis().getValidChassisMoves();

      chassisMoves.add(0, "none");
      chassis.setModel(new DefaultComboBoxModel(chassisMoves));

      power.addItem("Low power");
      // power.addItem("Medium power");
      power.addItem("High power");

      chassis.addActionListener(this);
      amount.addActionListener(this);
      power.addActionListener(this);
      // delay.addActionListener(this);

      updateEnabled = true;
    }


    public Component getTableCellEditorComponent
    (
      JTable table,
      Object value,
      boolean isSelected,
      int row, int column
    )
    {
      // BDebug.print ("BSnapshotEditor getTableCellEditorComponent("+ value+ ")");

      snap = (BSnapshot) value;
      updateUI(snap);

      repaint();

      return container;
    }

    public Object getCellEditorValue()
    {
      //BDebug.print ("BSnapshotEditor getCellEditorValue() ["+ snap+ "]" );

      if (snap != null)
        updateSnapshot(snap);

      shield.raise();
      return snap;
    }


    protected void updateUI(BSnapshot s)
    {
      updateEnabled = false;

      Image img = s.getThumbnail();

      if (img != null)
      {
        icon.setImage(img);
        thumb.setIcon(icon);
      }
      else
        thumb.setIcon(null);

      BMove move = s.getChassisMove();

      if (move != null)
        chassis.setSelectedItem(move);
      else
        chassis.setSelectedIndex(0);

      if (!BGlobal.CONSTRUCTING_CHASSISBOT)
      {
        amount.setVisible(move != null);
        amount.setValue(move != null ? s.getMetersToMove() : 0f);
      }
      else
        amount.setValue(s.getMetersToMove());

      units.setText(s.getChassisMoveUnits(true));
      power.setSelectedIndex(s.getPowerLevel());
      // delay.setSelected(s.getDelayJoints());

      updateEnabled = true;
    }


    protected BMove getSelectedChassisMove()
    {
      Object o = chassis.getSelectedItem();
      return (o instanceof BMove) ? (BMove) o : null;
    }

    protected void updateSnapshot(BSnapshot s)
    {
      BMove move = getSelectedChassisMove();

      s.setChassisMove(move);
      s.setMetersToMove(amount.getFloatValue());
      s.setPowerLevel(power.getSelectedIndex());
      s.setDelayJoints(false);
      // s.setDelayJoints(delay.isSelected());

    }

    public void refresh()
    {
      snap = control.getSelectedSnapshot();

      if (snap != null)
        updateUI(snap);

      repaint();
    }


    public void actionPerformed(ActionEvent e)
    {
      // BDebug.print ("BSnapshotEditor action performed fired");

      if (updateEnabled)
      {
        updateSnapshot(snap);
        updateUI(snap);

        control.skipTo(snap);
      }
    }

    public void tableChanged(TableModelEvent e)
    {
      // BDebug.print ( "BSnapshotEditor table changed fired ");

      int row = e.getFirstRow();
      //int row = e.getLastRow();


      if (snap != null)
        if (control.getSnapshot(row) == snap)
          updateUI(snap);
        else
          stopEditing();
    }

    protected class SnapRefresher extends BAction
    {
      public SnapRefresher()
      {
        super("Refresh snapshot");
        setIcon(BIconCache.getHSB("icon_snap_refresh.gif"));
        setShortDescription("Re-capture this snapshot.");
      }

      public void actionPerformed(ActionEvent e)
        { control.recordOver(); }
    }


    protected void stopEditing()
    {
      // BDebug.print("BSnapshotEditor.stopEditing()");
      snapshots.removeEditor();
      shield.raise();
    }

    protected void startEditing(BSnapshot s)
    {
      // BDebug.print("BSnapshotEditor.startEditing()");
      int r = control.indexOf(s);
      snapshots.editCellAt(r, 0);
      shield.lower();

      // snapshots.repaint();
      // editor.refresh();
    }

    protected class ShieldPanel extends BPanel
    {
      public ShieldPanel(BUIResources res, String key)
      { 
        super(res, key); 

        addMouseListener( new MouseAdapter()
          { public void mousePressed(MouseEvent e) { lower(); } });
      }
      
      public void raise()
        { setVisible(true); }
      
      public void lower()
        { setVisible(false); }
      
    }
  }


}