package roboforge.ui.components;

import roboforge.BDebug;
import roboforge.ui.dnd.*;
import roboforge.ui.components.table.*;

import java.awt.*;
import java.awt.dnd.*;
import java.awt.datatransfer.*;

import java.io.*;
import java.io.IOException;

import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.table.*;


public class BDragDropTable extends BTable
  implements
    DragSourceListener,
    DragGestureListener
{


  // Internal variables...
  // -------------------------------------------------------------

  protected DragSource dragSource  = null;
  protected DropTarget dropTarget  = null;

  protected Object[]   dragData    = null;
  protected int        dragIndex   = -1;
  protected int        dropIndex   = -1;

  protected static final DataFlavor OBJECT =
    ObjectSelection.createObjectFlavor();


  // Constructor...
  // -------------------------------------------------------------

  public BDragDropTable()
    { this((BUIResources) null, (String) null); }

  public BDragDropTable(BUIResources res, String key)
  {
    super(res, key);

    dragSource = new DragSource();
    dropTarget = new DropTarget(this, new DropListener());
    dragSource.createDefaultDragGestureRecognizer
      (this, DnDConstants.ACTION_COPY_OR_MOVE, this);
  }

  public BDragDropTable(Object[][] data, Object[] columnNames)
  {
    this((BUIResources) null, (String) null);

    BTableModel model =
      new BTableModel(data, columnNames);

    setModel(model);
  }


  // Drag 'n Drop methods...
  // -------------------------------------------------------------


  protected boolean acceptDrop(Object o, int i)
    // Determines what objects can be dropped on this list.
  {
    return true;
  }

  protected boolean acceptDrag(Object o, int i)
    // Determines what objects can be dragged off this list.
  {
    boolean isObjectArray = (o instanceof Object[]);
    boolean isNotEditing  = !isEditing();

    return (isObjectArray && isNotEditing);
  }


  public void dragGestureRecognized(DragGestureEvent e)
    // a drag gesture has been initiated
  {
    int r = getSelectedRow();

    TableModel model = getModel();

    if (r != -1)
      dragData = new Object[model.getColumnCount()];
    else
      dragData = null;

    if ((dragData != null) && acceptDrag(dragData, r))
    {
      System.out.println("Drag data: "+ dragData);

      // Reset dropIndex...
      dropIndex = -1;
      dragIndex = r;

      for (int i=0; i<dragData.length; i++)
        dragData[i] = model.getValueAt(r, i);

      ObjectSelection o = new ObjectSelection(dragData);

      Cursor cursor = null;
      switch (e.getDragAction())
      {
        case DnDConstants.ACTION_COPY:
          cursor = DragSource.DefaultCopyDrop;
          break;

        case DnDConstants.ACTION_MOVE:
          cursor = DragSource.DefaultMoveDrop;
          break;
      }

      dragSource.startDrag (e, cursor, o, this);
    }
    else
      { System.out.println( "nothing was selected");  }
  }

  public void dragDropEnd (DragSourceDropEvent e)
    // goes to DragSourceListener, informing it that dragging has ended.
  {
    System.out.println("DragDropEnd: success = "+ e.getDropSuccess());

    if (e.getDropSuccess())
      if (e.getDropAction() == DnDConstants.ACTION_MOVE)
        extractDragData();

    // Reset indices...
    dropIndex = -1;
    dragIndex = -1;
  }

  public void dragEnter (DragSourceDragEvent e)
    // goes to DragSourceListener, informing it that dragging has
    // entered the DropSite.
  {
    // System.out.println( "dragEnter");
  }

  public void dragExit (DragSourceEvent e)
    // this message goes to DragSourceListener, informing it that the
    // dragging has exited the DropSite.
  {
    // System.out.println( "dragExit");
  }

  public void dragOver (DragSourceDragEvent e)
    // this message goes to DragSourceListener, informing it that dragging
    // is currently occuring over the DropSite
  {
    // System.out.println( "dragExit");
  }

  public void dropActionChanged (DragSourceDragEvent e)
    // is invoked when the user changes the dropAction
  {
    // System.out.println( "dropActionChanged");
  }


  // -------------------------------------------------------------
  // Drop Target Listener Class...
  // -------------------------------------------------------------

  protected class DropListener implements DropTargetListener
  {

    public void dragEnter (DropTargetDragEvent e)
      // is invoked when you are dragging over the DropSite
    {
      // debug messages for diagnostics
      System.out.println( "dragEnter");
      e.acceptDrag(DnDConstants.ACTION_COPY_OR_MOVE);
    }

    public void dragExit (DropTargetEvent e)
      // is invoked when you are exit the DropSite without dropping
    {
      System.out.println("dragExit");
    }

    public void dragOver (DropTargetDragEvent e)
      // is invoked when a drag operation is going on
    {
      // System.out.println( "dragOver");
    }

    public void drop (DropTargetDropEvent e)
      // a drop has occurred
    {
      System.out.println("DROP DETECTED");

      try
      {
        Transferable t = e.getTransferable();
        if (t.isDataFlavorSupported(OBJECT))
        {
          Object o = t.getTransferData(OBJECT);
          int i = rowAtPoint(e.getLocation());

          if (acceptDrop(o, i))
          {
            e.acceptDrop(DnDConstants.ACTION_COPY_OR_MOVE);
            insertDropData(o, i);
            e.getDropTargetContext().dropComplete(true);
          }
          else
            e.rejectDrop();
        }
        else
          e.rejectDrop();
      }
      catch (IOException ex)
      {
        System.err.println("Exception: " + ex.getMessage());
        e.rejectDrop();
      }
      catch (UnsupportedFlavorException ex)
      {
        System.err.println("Exception: " + ex.getMessage());
        e.rejectDrop();
      }
    }

    public void dropActionChanged ( DropTargetDragEvent e )
      // is invoked if the use modifies the current drop gesture
    { }

  }


  // Add/Remove methods...
  // -------------------------------------------------------------

  public void insertDropData(Object o, int i)
  {
    DefaultTableModel model = (DefaultTableModel) getModel();

    dropIndex = i;

    // BDebug.print("insertDropData("+p.x+","+p.y+"):  PRE: drag= "+dragIndex+". drop= "+dropIndex);

    // Past end of table...
    if (dropIndex < 0)
      dropIndex = model.getRowCount();

    // handle local drag n' drop...
    else if (dragIndex >= 0)
    {
      if (dragIndex < dropIndex)
        dropIndex++;

      else if (dragIndex > dropIndex)
        dragIndex++;
    }

    // BDebug.print("insertDropData("+p.x+","+p.y+"): POST: drag= "+dragIndex+". drop= "+dropIndex);

    model.insertRow(dropIndex, (Object[]) o);

    getSelectionModel().setSelectionInterval(dropIndex, dropIndex);
  }

  public void extractDragData()
  {
    // BDebug.print("removeDropData(): index = "+dropIndex+". data = "+o);
    DefaultTableModel model = (DefaultTableModel) getModel();

    // Do the extraction...
    model.removeRow(dragIndex);
  }



}