package roboforge.ui.components;

import roboforge.BGlobal;
import roboforge.math.BMath;
import roboforge.util.graphics.BColors;

import java.awt.*;

import javax.swing.*;
import javax.vecmath.*;


public class BGrid2D extends BPanel
{

  // Class variables...
  // -------------------------------------------------------------

  protected static final double DEFAULT_METERS_PER_MAJOR = 5.0;
  protected static final int    DEFAULT_MINORS_PER_MAJOR = 5;

  protected static final double SMALLEST_SCALE           = 0.25;
  protected static final double LARGEST_SCALE            = 2.0;


  // Member variables...
  // -------------------------------------------------------------

  protected double  metersPerMajor;
  protected int     minorsPerMajor;

  protected double  xm, ym;
    // View location [meters].

  protected double  wm, hm;
    // Default view width/height, in meters.

  protected double scale;
    // View scale factor.

  protected Color minorColor =
    BGlobal.colors.getDarkestShade();

  protected Color majorColor =
    BColors.blendHSB(
      BGlobal.colors.getDarkestShade(),
      BGlobal.colors.getDarkerShade(), 0.5f);

  protected Color borderColor =
    BGlobal.colors.getDarkerShade();


  // Derived variables...
  // -------------------------------------------------------------

  protected double  metersPerMinor;
  protected double  pelsPerXMeter;
  protected double  pelsPerYMeter;
  protected double  pelsPerXMinor;
  protected double  pelsPerYMinor;
  protected double  numXMinors;
  protected double  numYMinors;
  protected double  metersPerXPel;
  protected double  metersPerYPel;


  protected double  tx, ty;
    // Top-left corner of view.

  protected int wp, hp;
    // View size, in pels.

  protected double ws, hs;
    // Scaled view width/height, in meters.

  protected Stroke strokeInMeters;


  // Constructors...
  // -------------------------------------------------------------

  public BGrid2D(BUIResources res, String key)
  {
    super(res, key);

    wm = hm        = res.getDouble(key+ ".meters");
    metersPerMajor = res.getDouble(key+ ".metersPerMajor");
    minorsPerMajor = res.getInt(key+ ".minorsPerMajor");
    metersPerMinor = metersPerMajor / minorsPerMajor;

    this.xm        = 0;
    this.ym        = 0;
    this.scale     = 1;

    calculateMetrics();
  }



  // Accessor methods...
  // -------------------------------------------------------------

  public void setViewDefaultSize(double wm, double hm)
  {
    this.wm = wm;
    this.hm = hm;
    calculateMetrics();

    repaint();
  }

  public void setViewScale(double scale)
  {
    if (scale < SMALLEST_SCALE)
      this.scale = SMALLEST_SCALE;
    else if (scale > LARGEST_SCALE)
      this.scale = LARGEST_SCALE;
    else
      this.scale = scale;

    calculateMetrics();

    repaint();
  }

  public void setViewLocation(Point2d p)
  {
    this.xm = p.x;
    this.ym = p.y;
    calculateMetrics();

    repaint();
  }

  protected void calculateMetrics()
  {
    ws = wm / scale;
    hs = hm / scale;

    tx = xm - (ws / 2);
    ty = ym - (hs / 2);

    Dimension d = getSize();
    wp = d.width;
    hp = d.height;

    numXMinors = ws / metersPerMinor;
    numYMinors = hs / metersPerMinor;

    pelsPerXMeter = wp / ws;
    pelsPerYMeter = hp / hs;
    metersPerXPel = ws / wp;
    metersPerYPel = hs / hp;

    pelsPerXMinor = wp / numXMinors;
    pelsPerYMinor = hp / numYMinors;

    strokeInMeters =
      new BasicStroke((float) metersPerXPel);
  }


  public Point2d pelsToMeters(Point2d p, Point2d m)
  {
    m.x = tx + (p.x / pelsPerXMeter);
    m.y = ty + (p.y / pelsPerYMeter);

    return m;
  }

  public Point2d metersToPels(Point2d m, Point2d p)
  {
    p.x = (m.x - tx) * pelsPerXMeter;
    p.y = (m.y - ty) * pelsPerYMeter;

    return p;
  }

  public double pelsToXMeters(double pels)
    { return pels / pelsPerXMeter; }

  public double pelsToYMeters(double pels)
    { return pels / pelsPerYMeter; }

  public double metersToXPels(double m)
    { return m * pelsPerXMeter; }

  public double metersToYPels(double m)
    { return m * pelsPerYMeter; }


  // Implementation methods...
  // -------------------------------------------------------------

  public void reshape(int x, int y, int w, int h)
  {
    super.reshape(x, y, w, h);
    calculateMetrics();
  }


  public void paint(Graphics g)
  {
    super.paint(g);

    g.setColor(borderColor);
    g.drawRect(0, 0, wp-1, hp-1);
    g.setClip(1, 1, wp-2, hp-2);

    double fx = Math.ceil(tx / metersPerMinor);
    double fy = Math.ceil(ty / metersPerMinor);

    double sx = (fx * metersPerMinor - tx) * pelsPerXMinor;
    double sy = (fy * metersPerMinor - ty) * pelsPerYMinor;

    // ....

    double x = sx;
    double y = sy;

    int bx = (int) fx;
    int by = (int) fy;
    int nx = (int) (fx + numXMinors);
    int ny = (int) (fy + numYMinors);

    int cx, cy, n;


    // Minor gridlines...

    g.setColor(minorColor);

    for (int i=bx; i<=nx; i++)
    {
      cx = (int) x;

      if ((i % minorsPerMajor) != 0)
        g.drawLine(cx, 0, cx, hp);

      x += pelsPerXMinor;
    }

    for (int i=by; i<=ny; i++)
    {
      cy = (int) y;

      if ((i % minorsPerMajor) != 0)
        g.drawLine(0, cy, wp, cy);

      y += pelsPerYMinor;
    }


    // Major gridlines...

    g.setColor(majorColor);

    x = sx;
    y = sy;

    for (int i = bx; i<=nx; i++)
    {
      cx = (int) x;

      if ((i % minorsPerMajor) == 0)
        g.drawLine(cx, 0, cx, hp);

      x += pelsPerXMinor;
    }

    for (int i = by; i<=ny; i++)
    {
      cy = (int) y;

      if ((i % minorsPerMajor) == 0)
        g.drawLine(0, cy, wp, cy);

      y += pelsPerYMinor;
    }
  }

}

