package roboforge;

import java.io.*;
import java.awt.*;
import javax.media.j3d.*;
import javax.vecmath.*;

import roboforge.sound.*;
import roboforge.arenaSystems.*;



public class BCollisionRecord implements Serializable, BArenaEvent
{

  static final long serialVersionUID = -1L;
  
  BComponent armComponent;
  transient BShape3D armShape;

  BComponent triggerComponent;
  transient BShape3D triggerShape;

  private transient BSound mySound;
  private int roundedMomentum;

  private static float MOMENTUM_BOTTOM = 50f; //40f;
  private static float MOMENTUM_TOP = 700f; //800f;
  private static int STATE_MODIFIER_MAX = 1500;

  private Point3d collisionPoint;


  private boolean nullified = false;   // nullified indicates that the collision is not a valid collision.
                                                 // if this flag is set, Damage and resultant moves will not be calculated from this collision, but rollback will occur as normal
  private transient boolean contacting = false;



  public BCollisionRecord(BComponent newArmComponent, BArenaObject newTriggerObject, BComponent newTriggerComponent)
  {
    set(newArmComponent, newTriggerComponent);
    nullified = false;

  }

  public void setCollisionMomentum(float m)   // called from BSimulation.processCollisions
  { roundedMomentum = java.lang.Math.round(m); }

  protected BComponent getArmComponent()
  { return armComponent; }

  protected BComponent getTriggerComponent()
  { return triggerComponent; }

  public void set(BComponent newArmComponent, BComponent newTriggerComponent)
  {
    armComponent = newArmComponent;
    triggerComponent = newTriggerComponent;

    if (armComponent.getCollisionCacheTime() > triggerComponent.getCollisionCacheTime() )
      collisionPoint =  armComponent.getCollisionPoint();
    else
      collisionPoint =  triggerComponent.getCollisionPoint();


  }


  public Point3d getCollisionPoint()
  {
    return  collisionPoint;
  }

  public void setNullified(boolean b)
  { nullified = b ;}

  public boolean getNullified()
  { return nullified ;}



  public String toString()
  {
    return "BCollisionRecord " + armComponent + " and " + triggerComponent + ":: nullified :: "+ getNullified() ;
  }


  public int getRoundedMomentum()
    { return roundedMomentum; }

  public float getCollisionIntensity()
  {
    // returns a float in [0..1], representing how
    // intense this collision is (i.e. how much momentum is involved.)

    if (roundedMomentum >= MOMENTUM_TOP)
      return 1f;
    else if (roundedMomentum <= MOMENTUM_BOTTOM)
      return 0f;
    else
      return (roundedMomentum - MOMENTUM_BOTTOM) /
        (MOMENTUM_TOP - MOMENTUM_BOTTOM);
  }


  public void initSound()
  {
    float gain = 0.25f + (getCollisionIntensity() * 0.75f);

    if (gain > 0.25f)
    {
       mySound = BComponentMaterial.getCollisionSound(armComponent.getMaterialType(), triggerComponent.getMaterialType()     ) ;
       mySound.setPriority(BSound.PRIORITY_MIN);
       mySound.setVolumeRatio(gain);
    }
  }

  public void playSound()
  {
  // BDebug.print ("initSound play " + roundedMomentum);

    if (mySound != null)
      //if (!mySound.isRunning())
      if ( System.currentTimeMillis() - mySound.getStartTime() > 250 )
        mySound.play();

  }

  public boolean useSameSound(BCollisionRecord other)
  {
    if (!BGlobal.USE_SOUND)
      return true;
    if ( (mySound == null) || (other.mySound == null)  )
      return false;


    return mySound.equals(other.mySound);

  }


  public boolean containsBot (BBot b)
  {
    return (armComponent.getBot() == b) || (triggerComponent.getBot() == b);
  }

  private static float  COL_TRAN = 0.4f;
  private static Color  COL_HIGHLIGHT_COLOR = new Color(255,0,0);
  
  private static BComponentAppearance COL_HIGHLIGHT = new BComponentAppearance(
    COL_HIGHLIGHT_COLOR,
    BComponentAppearance.DEFAULTDEPTH,
    0f,
    null,
    COL_TRAN);

  public void setHighlight(boolean set)
  {
    if (set)
    {
      // the shape may be live if a user removes components in the UI


      if (armComponent.getLiveNode() != null)
       armComponent.setHighlight(COL_HIGHLIGHT);
     // COL_HIGHLIGHT.setTexture(triggerComponent.getTextureName());
      if (triggerComponent.getLiveNode() != null)
        triggerComponent.setHighlight(COL_HIGHLIGHT);
    }
    else
    {
      if (armComponent.getLiveNode() != null)
        armComponent.setNormalAppearance();
      if (triggerComponent.getLiveNode() != null)
        triggerComponent.setNormalAppearance();
    }

  }



  public BCollisionRecord getBestCollision(BCollisionRecord other, BComponent commonComponent, javax.media.j3d.Locale myLocale)
  {
   // BDebug.print ("BCollisionRecord.getBestCollision entered ");

    BComponent myNotC = null;
    BComponent otherNotC = null;

    if( commonComponent == this.armComponent )
      myNotC = this.triggerComponent;
    else
      myNotC = this.armComponent;

    if( commonComponent == other.armComponent )
      otherNotC = other.triggerComponent;
    else
      otherNotC = other.armComponent;



    // best intersect
    BComponent bestIntersect = getBestCollision(commonComponent, myNotC, otherNotC,  myLocale);

    if ( (bestIntersect ==  this.armComponent) || (bestIntersect ==  this.triggerComponent)  )
    {
      return this;
    }
    else
      return other;


  }

  private static Transform3D intersect_Tran = new Transform3D();
  private static Vector3d intersect_Vec = new Vector3d();

  private static Point3d intersect_startPointCenter = new Point3d();
  private static Point3d intersect_endPointCenter = new Point3d();


  private boolean intersect( BComponent startPoint, BComponent endPoint, BComponent intTest , javax.media.j3d.Locale myLocale)
  {
    if ( startPoint == null || endPoint == null || intTest == null )
    {
      BDebug.print("BCollisionRecord.intersect: NULL component detected "+startPoint+" - "+endPoint+" - "+intTest);
      return false;
    } else if ( startPoint.getArbitraryShape_live() == null || endPoint.myShapesNode == null ) {
      BDebug.print("BCollisionRecord.intersect: NULL shape detected "+startPoint.getArbitraryShape_live()+" - "+endPoint.myShapesNode);
      return false;
    } else {

      startPoint.getArbitraryShape_live().getLocalToVworld(intersect_Tran);
      intersect_Tran.get(intersect_Vec );
      intersect_endPointCenter.set(intersect_Vec);

      endPoint.myShapesNode.getLocalToVworld(intersect_Tran);
      intersect_Tran.get(intersect_Vec );
      intersect_startPointCenter.set(intersect_Vec);

      PickSegment seg = new PickSegment(intersect_startPointCenter, intersect_endPointCenter);

      return BCollision.intersect(seg, intTest,  myLocale);
    }
  }




  private BComponent getBestCollision(BComponent commonComponent, BComponent a, BComponent b, javax.media.j3d.Locale myLocale)
  {
    //BDebug.print ("BCollisionRecord testing " + a + " and " + b + " common component " + commonComponent);
    BComponent best;

    if (b.getComponentType() == BWeapon.COMPONENT_TYPE)
        best = b;
    else if (a.getComponentType() == BWeapon.COMPONENT_TYPE)
        best = a;
    else if (b.getComponentType() == BShielding.COMPONENT_TYPE)
        best = b;
    else if (a.getComponentType() == BShielding.COMPONENT_TYPE)
        best = a;
    else if (intersect(commonComponent, a, b,  myLocale) )
    {
      best = b;
    //  BDebug.print ("BCollisionRecord intersect best =  " + best);
    }
    else if (intersect(commonComponent, b, a,  myLocale) )
    {
      best = a;
     // BDebug.print ("BCollisionRecord intersect best =  " + best);
    }
    else
    {

      if (b.getComponentType() == BCpu.COMPONENT_TYPE)
        best = a;
      else if (a.getComponentType() == BCpu.COMPONENT_TYPE)
        best = b;
      else if (b.getTreeNode() == null)
        best = a;
      else if (a.getTreeNode() == null)
        best = b;
      else if (b.getTreeNode().isNodeDescendant(a.getTreeNode()) )
        best = a;
      else if (a.getTreeNode().isNodeDescendant(b.getTreeNode()) )
        best = b;

      // if all else fails choose the largest one
      else if ( b.getVolume() > a.getVolume() )
        best = b;
      else
        best = a;
    }

    return best;


  }





  public BComponent getCommonComponent(BCollisionRecord other)
  {
    if (other.armComponent == this.armComponent)
      return  this.armComponent;
    else if  (other.triggerComponent == this.armComponent)
      return  this.armComponent;
    else if  (other.armComponent == this.triggerComponent)
      return  this.triggerComponent;
    else if  (other.triggerComponent == this.triggerComponent)
      return  this.triggerComponent;
    else return null;

  }



  public boolean equals(Object obj)
  {
      boolean e =  false;

      if (obj instanceof BCollisionRecord)
      {
        BCollisionRecord other = (BCollisionRecord) obj;

        if ( (armComponent == other.armComponent) && (triggerComponent == other.triggerComponent) )
          e = true;
        else if ( (armComponent == other.triggerComponent) && (triggerComponent == other.armComponent) )
          e = true;


      }

      return e;
  }


  public boolean isInContact()
    { return contacting; }

  public void setInContact(boolean b)
  {
    if (b)
      setNullified(true);

    contacting = b;

  }

  // Crowd Event momentum

  public int getStateModifier()  // the amount to modify the internal crowdState By
  { return java.lang.Math.min(roundedMomentum, STATE_MODIFIER_MAX);  }

  public String getArenaEventType()
  { return BArenaEvent.COLLISION ; }




}