package roboforge.ai;

import roboforge.BBot;

import roboforge.btree.*;
import roboforge.BDebug;

import roboforge.ai.decisions.*;
import roboforge.ai.expressions.*;
import roboforge.ai.memory.*;
import roboforge.ai.worldstate.*;
import roboforge.BMoveList;
import roboforge.BZones;

import java.io.Serializable;
import java.util.*;
import javax.swing.tree.*;
import javax.swing.event.*;

import roboforge.ai.expressions.parser.*;
import roboforge.botanalyst.*;



public class BDecisionTreeAI implements BAI, Serializable
{

  // Serialization version Unique ID
  // -----------------------------------------------------------
  static final long serialVersionUID = -1L;



  // Internal variables...
  // -----------------------------------------------------------

  // The bot that this AI controls.
  protected BBot bot;

  // Reference to the AI's decision tree.
  protected BSerializableTree tree;

  // The AI's memory bank - a collection of memory items.
  protected BMemoryBank memory;



  // 'Temporary' variables...
  // -----------------------------------------------------------

  // An enumeration of visitable nodes in the decision tree.
  protected transient Enumeration visitable;

  // Root node of this AI's decision tree.
  protected transient BRootNode  root;

  // Bot's perceived world state - updated in execute().
  protected transient BWorldState world;

  protected transient boolean fullCycleStarted;
  protected transient Hashtable visitedNodes;

  protected transient BExpressionParser parser;

  //static int STARTING_NODES = 10;


  // Constructor.
  // -----------------------------------------------------------

  public BDecisionTreeAI(BBot bot)
  {

    this.bot    = bot;
    this.memory = new BMemoryBank();
    this.root   = new BRootNode();
    this.tree   = new BSerializableTree(root);

  }

  
  // Initialization
  // -----------------------------------------------------------

  public void init()
  {
  //  BDebug.print ("BDecisionTree AI.init");
    root  = (BRootNode) tree.getRoot();
    world = bot.getWorldState();

    memory.init();
    memory.addListDataListener(new MemoryPruneListener() );
    root.init();

    // Setting working variables
    visitable = null;

    initPrimaryKeys();

    visitedNodes = new Hashtable();


  }


  public void initPrimaryKeys()
  {
    // WARNING, changing this algorithm may change the way the battle player relates historical fights
    Enumeration allAi = ((DefaultMutableTreeNode)getRoot()).breadthFirstEnumeration();
    int i = 0;

    while(allAi.hasMoreElements())
    {
      Object next = allAi.nextElement();
      if (next instanceof BDecisionTreeNode)
      {
        ((BDecisionTreeNode)next).setPrimaryKey(new Integer(i));
        i ++;

      }
    }

  }

  public void updateVisited(Vector visited)
  {
    if (getRoot() == null)
      return;


    Enumeration allAi = ((DefaultMutableTreeNode)getRoot()).breadthFirstEnumeration();

    while(allAi.hasMoreElements())
    {
      Object next = allAi.nextElement();
      if (next instanceof BDecisionTreeNode)
      {
         BDecisionTreeNode current = (BDecisionTreeNode)next;
         current.setVisited(visited.indexOf(current.getPrimaryKey()) != -1 );

      }
    }

    //BDebug.printTime(" BDecisiontTreeAI.updateVisited " );


  }


  // Execution method...
  // -----------------------------------------------------------

  public void execute(int operations)
  {
    fullCycleStarted = false;

    if (visitable == null)
      startDecisionCycle();

    BDecisionTreeNode node;

    for (int i=0; i<operations; i++)
    {

      if (cycleFinished())
      {
        if (fullCycleStarted)
          return;

        startDecisionCycle();

        if (cycleFinished())
          return;
      }

      node = nextNodeInCycle();
      node.visit();
      visitedNodes.put(node.getPrimaryKey(), node.getPrimaryKey() );
    }
  }


    public Vector getAndCleanVisitedNodes()
    {
      Vector returnVector =  new Vector(visitedNodes.values() );
      visitedNodes.clear();

      return returnVector;

    }

     //visitedNodes.add(a.getAndCleanVisitedNodes() );



  // Execution support methods...
  // -------------------------------------------------------------


  protected void startDecisionCycle()
  {
    // Reset the decision flow...
    visitable = root.visitableSubtree();

    // Update our perceived world-state.
    world.evaluate();

    fullCycleStarted = true;
  }

  protected BDecisionTreeNode nextNodeInCycle()
    { return (BDecisionTreeNode) visitable.nextElement(); }

  protected boolean cycleFinished()
    { return !visitable.hasMoreElements(); }




  // Accessor methods...
  // -----------------------------------------------------------

  public BBot getBot()
    { return bot; }

  public void setBot(BBot newBot)
  {
    if (newBot != null)
      root  = (BRootNode) tree.getRoot();

    bot = newBot;

    Enumeration allAi = ((DefaultMutableTreeNode)getRoot()).breadthFirstEnumeration();

    while(allAi.hasMoreElements())
    {
      Object next = allAi.nextElement();
      if (next instanceof BDecisionTreeNode)
        ((BDecisionTreeNode)next).setActor(newBot);
    }

    if (newBot != null)
      init();
  }

  // to allow a shared instance for checking items
  public BExpressionParser getParser()
  {
    if (parser == null)
      parser = new  BExpressionParser(this);

    return parser;
  }

  public void prune()
  {
    Enumeration allAi = ((DefaultMutableTreeNode)getRoot()).breadthFirstEnumeration();

    while(allAi.hasMoreElements())
    {
      Object next = allAi.nextElement();
      ((BDecisionTreeNode)next).prune();
    }

  }


  public Vector getReferencedZoneNames()
  {
      Vector zones = new Vector();
      Enumeration allAi = ((DefaultMutableTreeNode)getRoot()).breadthFirstEnumeration();

      while(allAi.hasMoreElements())
      {
        Object next = allAi.nextElement();
        if (next instanceof BDecisionNode )
        {
          BDecisionNode node = (BDecisionNode) next;
          Vector nodeZones = node.getQuestionedZones();

          if (nodeZones != null)
            zones.addAll(nodeZones);

        }

      }


      return zones;


  }

  public Vector getReferencedMoves()
  {
      Vector moves = new Vector();
      Enumeration allAi = ((DefaultMutableTreeNode)getRoot()).breadthFirstEnumeration();

      while(allAi.hasMoreElements())
      {
        Object  next = allAi.nextElement();
        if (next instanceof BMoveActionNode)
        {
          BMoveActionNode node = (BMoveActionNode) next;
          if (node.getMoveAction() != null)
            moves.add(node.getMoveAction() );
        }

      }


      return moves;

  }

  public void validate(Vector warnings, boolean onlyFatal)
  {

    if (!onlyFatal)
    {
    // invalid nodes
      Enumeration allAi = ((DefaultMutableTreeNode)getRoot()).breadthFirstEnumeration();
      while(allAi.hasMoreElements())
      {
        BDecisionTreeNode  next = (BDecisionTreeNode)allAi.nextElement();
        if (!next.isValid())
          warnings.add(new BBotWarning( "AI node: " + next.getName() + " is invalid." ,
                                    "If this is a decision node, the Bot's AI will not look at any nodes that are below this node. If this is an 'action' (The green colored icons) node then the action will not be performed.") );

      }

      // non-referenced Moves
      Vector refM = getReferencedMoves();
      BMoveList mList = bot.getMoveList();
      for (int i = mList.size()-1; i >= 0; i -- )
        if (refM.indexOf(mList.getAt(i)) == -1)
          if ( ! ( mList.getAt(i) == mList.getStartMove()  || mList.getAt(i) == mList.getVictoryMove() ) )
            warnings.add(new BBotWarning( "Move: " + mList.getAt(i).toString() + " is not used in this bots AI." ,
                                      "") );


      Vector refZ = getReferencedZoneNames();
      //BDebug.print ("BDecisionTreeAI refZ" + refZ);
      BZones zList = bot.getZones();
      
      for (int i = zList.size()-1; i >= 0; i -- )
        if (refZ.indexOf( zList.getAt(i).toString() ) == -1)
          warnings.add(new BBotWarning( "Zone: " + zList.getAt(i).toString() + " is not used in this bots AI." ,
                                    "") );


     




    }


  }




  public BWorldState getWorldState()
    { return bot.getWorldState(); }

  public void setRoot(BRootNode root)
  {
    this.root = root;
    tree.setRoot(root);
  }

  public BRootNode getRoot()
    { return root; }

  public BMemoryBank getMemoryBank()
    { return memory; }



  class MemoryPruneListener implements ListDataListener
  {

      public void intervalAdded(ListDataEvent e)
        {}

      public void intervalRemoved(ListDataEvent e)
      {
        prune();
      }

      public void contentsChanged(ListDataEvent e)
        {}
  }



}
