package roboforge.ai.expressions.parser;

import roboforge.ai.*;
import roboforge.ai.expressions.*;
import roboforge.ai.memory.*;
import roboforge.ai.values.*;

import roboforge.util.parser.*;

import java.io.*;
import java.util.HashMap;


public class BExpressionLexer implements
  BLexicalAnalyzer, BExpressionIDs
{

  // Internal variables...
  // -------------------------------------------------------------

  protected StreamTokenizer tokenizer;
    // Underlying tokenizer - handles word / number recognition.

  protected BSymbol current;
    // The current symbol.

  protected boolean symbolPushed;
    // Has a symbol been pushed back onto the stack?

  protected BDecisionTreeAI ai;
    // The AI context for this expression.

  protected BMemoryBank bank;
    // Lookup object for identifiers.



  // Class variables...
  // -------------------------------------------------------------

  protected static HashMap reserved = new HashMap(20);
    // Collection of reserved words.
    // Each reserved word has its own token ID.
  static
  {
    reserved.put("yes",   new BSymbol(CONSTANT, BConstants.TRUE));
    reserved.put("no",    new BSymbol(CONSTANT, BConstants.FALSE));
    reserved.put("true",  new BSymbol(CONSTANT, BConstants.TRUE));
    reserved.put("false", new BSymbol(CONSTANT, BConstants.FALSE));
    reserved.put("and",   new BSymbol(AND,      new BAnd()));
    reserved.put("or",    new BSymbol(OR,       new BOr()));
    reserved.put("not",   new BSymbol(NOT,      new BNot()));
    reserved.put("pi",    new BSymbol(CONSTANT, BConstants.PI));
  }



  // Constructor...
  // -------------------------------------------------------------

  public BExpressionLexer(BDecisionTreeAI ai)
  {
    this.ai   = ai;
    this.bank = ai.getMemoryBank();
  }


  // Input method...
  // -------------------------------------------------------------

  public void setInput(Reader reader)
  {
    // StringReader reader = new StringReader(text);

    tokenizer = new StreamTokenizer(reader);
    configTokenizer();

    current       = new BSymbol(BSymbol.NOTHING);
    symbolPushed  = false;
  }


  // Initialization methods...
  // -------------------------------------------------------------

  protected void configTokenizer()
  {
    tokenizer.resetSyntax();

    /*
    tokenizer.ordinaryChar('+');
    tokenizer.ordinaryChar('-');
    tokenizer.ordinaryChar('*');
    tokenizer.ordinaryChar('/');
    tokenizer.ordinaryChar('>');
    tokenizer.ordinaryChar('<');
    tokenizer.ordinaryChar('=');
    tokenizer.ordinaryChar('(');
    tokenizer.ordinaryChar(')');
    tokenizer.ordinaryChar('[');
    tokenizer.ordinaryChar(']');
    tokenizer.ordinaryChar('.');
    tokenizer.ordinaryChars(0, ' ');
    */

    tokenizer.lowerCaseMode(true);
  	tokenizer.parseNumbers();

  	tokenizer.quoteChar('"');

  	tokenizer.wordChars('a', 'z');
	  tokenizer.wordChars('A', 'Z');
	  tokenizer.wordChars('_', '_');

  	// tokenizer.whitespaceChars(0, ' ');
  }

  protected void configTokenizerForReferences()
  {
    tokenizer.resetSyntax();
  }


  // Accessor methods...
  // -------------------------------------------------------------

  public BSymbol nextSymbol() throws BLexicalException
    // return the next token in input stack.
  {
    if (symbolPushed)
    {
      symbolPushed = false;
      return current;
    }

    int token, id, next;
    Object value;

    try
    {
      token = tokenizer.nextToken();

      switch (token)
      {
        case StreamTokenizer.TT_EOF:
          id    = EOF;
          value = null;
          break;

        case StreamTokenizer.TT_NUMBER:
          id    = NUMBER;
          value = new BConstant(new BNumber(tokenizer.nval));
          break;

        case StreamTokenizer.TT_WORD:

          // Perform reserved-word lookup...
          BSymbol s = (BSymbol) reserved.get(tokenizer.sval);

          if (s != null)
          {
            id    = s.getID();
            value = ((BExpression) s.getValue()).clone();
          }
          else
          {
            id = IDENTIFIER;

            BMemoryItem item = bank.getItem(tokenizer.sval);

            if (item != null)
              value = new BMemoryRef(item);
            else
              value = new BConstant(new BString(tokenizer.sval));
              // TODO: Error goes here.
          }
          break;

        case '"':
          id    = STRING;
          value = new BConstant(new BString(tokenizer.sval));
          break;

        case '+':
          id    = PLUS;
          value = new BPlus();
          break;

        case '-':
          id    = MINUS;
          value = new BMinus();
          break;

        case '*':
          id    = TIMES;
          value = new BTimes();
          break;

        case '/':
          id    = DIVIDE;
          value = new BDiv();
          break;

        case '=':
          id    = EQ;
          value = new BEquals();
          break;

        case '(':
          id    = LPAREN;
          value = new BParenthesis();
          break;

        case ')':
          id    = RPAREN;
          value = null;
          break;

        case '>':
          next = tokenizer.nextToken();

          if (next == '=')
          {
            id    = GEQ;
            value = new BGreaterEqThan();
          }
          else
          {
            tokenizer.pushBack();
            id    = GTR;
            value = new BGreaterThan();
          }
          break;

        case '<':
          next = tokenizer.nextToken();

          if (next == '=')
          {
            id    = LEQ;
            value = new BLessEqThan();
          }
          else
          {
            tokenizer.pushBack();
            id    = LESS;
            value = new BLessThan();
          }
          break;

        case '!':
          next = tokenizer.nextToken();

          if (next == '=')
          {
            id    = NEQ;
            value = new BNotEquals();
          }
          else
          {
            tokenizer.pushBack();
            id    = NOT;
            value = new BNot();
          }
          break;

        case '[':
          configTokenizerForReferences();

          boolean closed = false;
          String key = "";
          char c;

          while (!closed)
          {
            next = tokenizer.nextToken();

            if (next == StreamTokenizer.TT_EOF)
              throw new BLexicalException("Unexpected end of reference: "+ key);

            closed = (next == ']');

            if (!closed && (next >= ' '))
              { c = (char) next; key += c; }
          }

          id    = REFERENCE;
          value = new BWorldItemRef(ai, key);

          configTokenizer();
          break;

        default:
          if ((token >= 0) && (token <= ' '))
          {
            consumeWhiteSpace();
            return nextSymbol();
          }
          else
            throw new BLexicalException("Unrecognized token: "+ (char) token);
      }

      current = new BSymbol(id, value);
      return current;
    }
    catch (IOException e)
      { throw new BLexicalException("Out of input."); }
  }

  public void pushSymbol()
    // push id token back onto input stack.
  {
    if (current.getID() != BSymbol.NOTHING)
      symbolPushed = true;
  }

  public BSymbol peekSymbol()
    { return current; }



  protected void consumeWhiteSpace() throws IOException
  {
    boolean skipping = true;

    int token;

    while (skipping)
    {
      token    = tokenizer.nextToken();
      skipping = (token >= 0) && (token <= ' ');
    }

    tokenizer.pushBack();
  }

}
