package roboforge.ai.expressions.parser;

import roboforge.ai.*;
import roboforge.ai.expressions.*;
import roboforge.ai.memory.*;
import roboforge.ai.values.*;
import roboforge.util.parser.*;

import java.io.*;
import java.util.Stack;


public class BExpressionParser extends BLALRParser
{

  // Constructor...
  // -------------------------------------------------------------

  public BExpressionParser(BDecisionTreeAI ai)
    { super(new BExpressionLexer(ai)); }


  // Accessor methods...
  // -------------------------------------------------------------

  public BExpression parse(String text) throws BParseException
  {
    // Trivially reject empty strings...
    if ((text == null) || text.trim().equals(""))
      return null;

    StringReader reader = new StringReader(text);

    super.parseInput(reader);

    BSymbol s = stack.pop();
    return (BExpression) s.getValue();
  }

  public BValue parseForValue(String text)
  {
    try
    {
      BExpression e = parse(text);
      return e.evaluate();
    }
    catch
      ( BParseException ex )
      { return BValue.UNDEFINED; }
  }


  // Rule classes...
  // -------------------------------------------------------------

  protected abstract class ExpressionRule extends Rule
  {
    protected String name;

    public ExpressionRule(String name)
      { super(0); this.name = name; }

    public String toString()
      { return name; }
  }

  protected class TerminalRule extends ExpressionRule
  {
    public TerminalRule(String name)
      { super(name); }

    public BSymbol apply() throws BParseException
      { return stack.pop(); }
  }

  protected class BinaryRule extends ExpressionRule
  {
    protected BBinaryOp   o;
    protected BExpression l;
    protected BExpression r;

    public BinaryRule(String name)
      { super(name); }

    public BSymbol apply() throws BParseException
    {
      r = (BExpression) stack.pop().getValue();
      o = (BBinaryOp)   stack.pop().getValue();
      l = (BExpression) stack.pop().getValue();

      o.setLeft(l);
      o.setRight(r);

      return BSymbol.nonTerminal("E", o);
    }
  }

  protected class UnaryRule extends ExpressionRule
  {
    protected BUnaryOp    o;
    protected BExpression e;

    public UnaryRule(String name)
      { super(name); }

    public BSymbol apply() throws BParseException
    {
      e = (BExpression) stack.pop().getValue();
      o = (BUnaryOp)    stack.pop().getValue();

      o.setOperand(e);
      return BSymbol.nonTerminal("E", o);
    }
  }

  protected class ParenthesisRule extends ExpressionRule
  {
    protected BUnaryOp    l;
    protected BExpression e;
    protected Object      r;

    public ParenthesisRule()
      { super("E -> ( E )"); }

    public BSymbol apply() throws BParseException
    {
      r = (Object)       stack.pop().getValue();
      e = (BExpression)  stack.pop().getValue();
      l = (BParenthesis) stack.pop().getValue();

      l.setOperand(e);
      return BSymbol.nonTerminal("E", l);
    }
  }


  // Table initialization...
  // -------------------------------------------------------------

  protected static final int NUM_TERMINALS = 21;
  protected static final int NUM_RULES     = 20;
  protected static final int NUM_STATES    = 37;

  protected void initRulesTable()
  {
    rules = new Rule[NUM_RULES];

    rules[ 0] = null;

    rules[ 1] = new BinaryRule("E -> E * E");
    rules[ 2] = new BinaryRule("E -> E / E");
    rules[ 3] = new BinaryRule("E -> E + E");
    rules[ 4] = new BinaryRule("E -> E - E");
    rules[ 5] = new BinaryRule("E -> E > E");
    rules[ 6] = new BinaryRule("E -> E < E");
    rules[ 7] = new BinaryRule("E -> E >= E");
    rules[ 8] = new BinaryRule("E -> E <= E");
    rules[ 9] = new BinaryRule("E -> E = E");
    rules[10] = new BinaryRule("E -> E <> E");
    rules[11] = new UnaryRule( "E -> not E");
    rules[12] = new BinaryRule("E -> E and E");
    rules[13] = new BinaryRule("E -> E or E");

    rules[14] = new ParenthesisRule();
    rules[15] = new TerminalRule("E -> <number>");
    rules[16] = new TerminalRule("E -> <string>");
    rules[17] = new TerminalRule("E -> <ident>");
    rules[18] = new TerminalRule("E -> <reference>");

    rules[19] = new TerminalRule("E -> <constant>");
  }

  protected void initGoToTable()
  {
    go_to = new int[NUM_STATES][1];

    go_to[ 0][0] = 8;
    go_to[ 3][0] = 36;
    go_to[ 7][0] = 34;
    go_to[ 9][0] = 33;
    go_to[10][0] = 32;
    go_to[11][0] = 31;
    go_to[12][0] = 30;
    go_to[13][0] = 29;
    go_to[14][0] = 28;
    go_to[16][0] = 27;
    go_to[17][0] = 26;
    go_to[18][0] = 25;
    go_to[19][0] = 24;
    go_to[20][0] = 23;
    go_to[21][0] = 22;
  }

  protected void initActionTable()
  {
    actions = new ParseAction[NUM_STATES][NUM_TERMINALS];

    Rule r0  = rules[ 0];
    Rule r1  = rules[ 1];
    Rule r2  = rules[ 2];
    Rule r3  = rules[ 3];
    Rule r4  = rules[ 4];
    Rule r5  = rules[ 5];
    Rule r6  = rules[ 6];
    Rule r7  = rules[ 7];
    Rule r8  = rules[ 8];
    Rule r9  = rules[ 9];
    Rule r10 = rules[10];
    Rule r11 = rules[11];
    Rule r12 = rules[12];
    Rule r13 = rules[13];
    Rule r14 = rules[14];
    Rule r15 = rules[15];
    Rule r16 = rules[16];
    Rule r17 = rules[17];
    Rule r18 = rules[18];
    Rule r19 = rules[19];


		actions[ 0][ 1]	= new Shift( 4);
		actions[ 0][ 2]	= new Shift( 5);
		actions[ 0][ 3]	= new Shift( 1);
		actions[ 0][ 4]	= new Shift( 6);
		actions[ 0][ 5]	= new Shift( 2);
		actions[ 0][ 6]	= new Shift( 7);
		actions[ 0][16]	= new Shift( 3);

		actions[ 1][ 0]	= new Reduce(r16);
		actions[ 1][ 7]	= new Reduce(r16);
		actions[ 1][ 8]	= new Reduce(r16);
		actions[ 1][ 9]	= new Reduce(r16);
		actions[ 1][10]	= new Reduce(r16);
		actions[ 1][11]	= new Reduce(r16);
		actions[ 1][12]	= new Reduce(r16);
		actions[ 1][13]	= new Reduce(r16);
		actions[ 1][14]	= new Reduce(r16);
		actions[ 1][15]	= new Reduce(r16);
		actions[ 1][17]	= new Reduce(r16);
		actions[ 1][18]	= new Reduce(r16);
		actions[ 1][19]	= new Reduce(r16);
		actions[ 1][20]	= new Reduce(r16);

		actions[ 2][ 0]	= new Reduce(r19);
		actions[ 2][ 7]	= new Reduce(r19);
		actions[ 2][ 8]	= new Reduce(r19);
		actions[ 2][ 9]	= new Reduce(r19);
		actions[ 2][10]	= new Reduce(r19);
		actions[ 2][11]	= new Reduce(r19);
		actions[ 2][12]	= new Reduce(r19);
		actions[ 2][13]	= new Reduce(r19);
		actions[ 2][14]	= new Reduce(r19);
		actions[ 2][15]	= new Reduce(r19);
		actions[ 2][17]	= new Reduce(r19);
		actions[ 2][18]	= new Reduce(r19);
		actions[ 2][19]	= new Reduce(r19);
		actions[ 2][20]	= new Reduce(r19);

		actions[ 3][ 1]	= new Shift( 4);
		actions[ 3][ 2]	= new Shift( 5);
		actions[ 3][ 3]	= new Shift( 1);
		actions[ 3][ 4]	= new Shift( 6);
		actions[ 3][ 5]	= new Shift( 2);
		actions[ 3][ 6]	= new Shift( 7);
		actions[ 3][16]	= new Shift( 3);

		actions[ 4][ 0]	= new Reduce(r15);
		actions[ 4][ 7]	= new Reduce(r15);
		actions[ 4][ 8]	= new Reduce(r15);
		actions[ 4][ 9]	= new Reduce(r15);
		actions[ 4][10]	= new Reduce(r15);
		actions[ 4][11]	= new Reduce(r15);
		actions[ 4][12]	= new Reduce(r15);
		actions[ 4][13]	= new Reduce(r15);
		actions[ 4][14]	= new Reduce(r15);
		actions[ 4][15]	= new Reduce(r15);
		actions[ 4][17]	= new Reduce(r15);
		actions[ 4][18]	= new Reduce(r15);
		actions[ 4][19]	= new Reduce(r15);
		actions[ 4][20]	= new Reduce(r15);

		actions[ 5][ 0]	= new Reduce(r17);
		actions[ 5][ 7]	= new Reduce(r17);
		actions[ 5][ 8]	= new Reduce(r17);
		actions[ 5][ 9]	= new Reduce(r17);
		actions[ 5][10]	= new Reduce(r17);
		actions[ 5][11]	= new Reduce(r17);
		actions[ 5][12]	= new Reduce(r17);
		actions[ 5][13]	= new Reduce(r17);
		actions[ 5][14]	= new Reduce(r17);
		actions[ 5][15]	= new Reduce(r17);
		actions[ 5][17]	= new Reduce(r17);
		actions[ 5][18]	= new Reduce(r17);
		actions[ 5][19]	= new Reduce(r17);
		actions[ 5][20]	= new Reduce(r17);

		actions[ 6][ 0]	= new Reduce(r18);
		actions[ 6][ 7]	= new Reduce(r18);
		actions[ 6][ 8]	= new Reduce(r18);
		actions[ 6][ 9]	= new Reduce(r18);
		actions[ 6][10]	= new Reduce(r18);
		actions[ 6][11]	= new Reduce(r18);
		actions[ 6][12]	= new Reduce(r18);
		actions[ 6][13]	= new Reduce(r18);
		actions[ 6][14]	= new Reduce(r18);
		actions[ 6][15]	= new Reduce(r18);
		actions[ 6][17]	= new Reduce(r18);
		actions[ 6][18]	= new Reduce(r18);
		actions[ 6][19]	= new Reduce(r18);
		actions[ 6][20]	= new Reduce(r18);

		actions[ 7][ 1]	= new Shift( 4);
		actions[ 7][ 2]	= new Shift( 5);
		actions[ 7][ 3]	= new Shift( 1);
		actions[ 7][ 4]	= new Shift( 6);
		actions[ 7][ 5]	= new Shift( 2);
		actions[ 7][ 6]	= new Shift( 7);
		actions[ 7][16]	= new Shift( 3);

		actions[ 8][ 0]	= new Accept();
		actions[ 8][ 8]	= new Shift(14);
		actions[ 8][ 9]	= new Shift(12);
		actions[ 8][10]	= new Shift(16);
		actions[ 8][11]	= new Shift(17);
		actions[ 8][12]	= new Shift(19);
		actions[ 8][13]	= new Shift(21);
		actions[ 8][14]	= new Shift( 9);
		actions[ 8][15]	= new Shift(20);
		actions[ 8][17]	= new Shift(13);
		actions[ 8][18]	= new Shift(10);
		actions[ 8][19]	= new Shift(18);
		actions[ 8][20]	= new Shift(11);

		actions[ 9][ 1]	= new Shift( 4);
		actions[ 9][ 2]	= new Shift( 5);
		actions[ 9][ 3]	= new Shift( 1);
		actions[ 9][ 4]	= new Shift( 6);
		actions[ 9][ 5]	= new Shift( 2);
		actions[ 9][ 6]	= new Shift( 7);
		actions[ 9][16]	= new Shift( 3);

		actions[10][ 1]	= new Shift( 4);
		actions[10][ 2]	= new Shift( 5);
		actions[10][ 3]	= new Shift( 1);
		actions[10][ 4]	= new Shift( 6);
		actions[10][ 5]	= new Shift( 2);
		actions[10][ 6]	= new Shift( 7);
		actions[10][16]	= new Shift( 3);

		actions[11][ 1]	= new Shift( 4);
		actions[11][ 2]	= new Shift( 5);
		actions[11][ 3]	= new Shift( 1);
		actions[11][ 4]	= new Shift( 6);
		actions[11][ 5]	= new Shift( 2);
		actions[11][ 6]	= new Shift( 7);
		actions[11][16]	= new Shift( 3);

		actions[12][ 1]	= new Shift( 4);
		actions[12][ 2]	= new Shift( 5);
		actions[12][ 3]	= new Shift( 1);
		actions[12][ 4]	= new Shift( 6);
		actions[12][ 5]	= new Shift( 2);
		actions[12][ 6]	= new Shift( 7);
		actions[12][16]	= new Shift( 3);

		actions[13][ 1]	= new Shift( 4);
		actions[13][ 2]	= new Shift( 5);
		actions[13][ 3]	= new Shift( 1);
		actions[13][ 4]	= new Shift( 6);
		actions[13][ 5]	= new Shift( 2);
		actions[13][ 6]	= new Shift( 7);
		actions[13][16]	= new Shift( 3);

		actions[14][ 1]	= new Shift( 4);
		actions[14][ 2]	= new Shift( 5);
		actions[14][ 3]	= new Shift( 1);
		actions[14][ 4]	= new Shift( 6);
		actions[14][ 5]	= new Shift( 2);
		actions[14][ 6]	= new Shift( 7);
		actions[14][16]	= new Shift( 3);

    /*
		actions[15][ 0]	= new Accept();
    */

		actions[16][ 1]	= new Shift( 4);
		actions[16][ 2]	= new Shift( 5);
		actions[16][ 3]	= new Shift( 1);
		actions[16][ 4]	= new Shift( 6);
		actions[16][ 5]	= new Shift( 2);
		actions[16][ 6]	= new Shift( 7);
		actions[16][16]	= new Shift( 3);

		actions[17][ 1]	= new Shift( 4);
		actions[17][ 2]	= new Shift( 5);
		actions[17][ 3]	= new Shift( 1);
		actions[17][ 4]	= new Shift( 6);
		actions[17][ 5]	= new Shift( 2);
		actions[17][ 6]	= new Shift( 7);
		actions[17][16]	= new Shift( 3);

		actions[18][ 1]	= new Shift( 4);
		actions[18][ 2]	= new Shift( 5);
		actions[18][ 3]	= new Shift( 1);
		actions[18][ 4]	= new Shift( 6);
		actions[18][ 5]	= new Shift( 2);
		actions[18][ 6]	= new Shift( 7);
		actions[18][16]	= new Shift( 3);

		actions[19][ 1]	= new Shift( 4);
		actions[19][ 2]	= new Shift( 5);
		actions[19][ 3]	= new Shift( 1);
		actions[19][ 4]	= new Shift( 6);
		actions[19][ 5]	= new Shift( 2);
		actions[19][ 6]	= new Shift( 7);
		actions[19][16]	= new Shift( 3);

		actions[20][ 1]	= new Shift( 4);
		actions[20][ 2]	= new Shift( 5);
		actions[20][ 3]	= new Shift( 1);
		actions[20][ 4]	= new Shift( 6);
		actions[20][ 5]	= new Shift( 2);
		actions[20][ 6]	= new Shift( 7);
		actions[20][16]	= new Shift( 3);

		actions[21][ 1]	= new Shift( 4);
		actions[21][ 2]	= new Shift( 5);
		actions[21][ 3]	= new Shift( 1);
		actions[21][ 4]	= new Shift( 6);
		actions[21][ 5]	= new Shift( 2);
		actions[21][ 6]	= new Shift( 7);
		actions[21][16]	= new Shift( 3);

		actions[22][ 0]	= new Reduce(r8);
		actions[22][ 7]	= new Reduce(r8);
		actions[22][ 8]	= new Reduce(r8);
		actions[22][ 9]	= new Reduce(r8);
		actions[22][14]	= new Reduce(r8);
		actions[22][15]	= new Reduce(r8);
		actions[22][17]	= new Shift(13);
		actions[22][18]	= new Shift(10);
		actions[22][19]	= new Shift(18);
		actions[22][20]	= new Shift(11);

		actions[23][ 0]	= new Reduce(r13);
		actions[23][ 7]	= new Reduce(r13);
		actions[23][ 8]	= new Shift(14);
		actions[23][ 9]	= new Shift(12);
		actions[23][10]	= new Shift(16);
		actions[23][11]	= new Shift(17);
		actions[23][12]	= new Shift(19);
		actions[23][13]	= new Shift(21);
		actions[23][14]	= new Shift( 9);
		actions[23][15]	= new Reduce(r13);
		actions[23][17]	= new Shift(13);
		actions[23][18]	= new Shift(10);
		actions[23][19]	= new Shift(18);
		actions[23][20]	= new Shift(11);

		actions[24][ 0]	= new Reduce(r7);
		actions[24][ 7]	= new Reduce(r7);
		actions[24][ 8]	= new Reduce(r7);
		actions[24][ 9]	= new Reduce(r7);
		actions[24][14]	= new Reduce(r7);
		actions[24][15]	= new Reduce(r7);
		actions[24][17]	= new Shift(13);
		actions[24][18]	= new Shift(10);
		actions[24][19]	= new Shift(18);
		actions[24][20]	= new Shift(11);

		actions[25][ 0]	= new Reduce(r1);
		actions[25][ 7]	= new Reduce(r1);
		actions[25][ 8]	= new Reduce(r1);
		actions[25][ 9]	= new Reduce(r1);
		actions[25][10]	= new Reduce(r1);
		actions[25][11]	= new Reduce(r1);
		actions[25][12]	= new Reduce(r1);
		actions[25][13]	= new Reduce(r1);
		actions[25][14]	= new Reduce(r1);
		actions[25][15]	= new Reduce(r1);
		actions[25][17]	= new Reduce(r1);
		actions[25][18]	= new Reduce(r1);
		actions[25][19]	= new Reduce(r1);
		actions[25][20]	= new Reduce(r1);

		actions[26][ 0]	= new Reduce(r6);
		actions[26][ 7]	= new Reduce(r6);
		actions[26][ 8]	= new Reduce(r6);
		actions[26][ 9]	= new Reduce(r6);
		actions[26][14]	= new Reduce(r6);
		actions[26][15]	= new Reduce(r6);
		actions[26][17]	= new Shift(13);
		actions[26][18]	= new Shift(10);
		actions[26][19]	= new Shift(18);
		actions[26][20]	= new Shift(11);

		actions[27][ 0]	= new Reduce(r5);
		actions[27][ 7]	= new Reduce(r5);
		actions[27][ 8]	= new Reduce(r5);
		actions[27][ 9]	= new Reduce(r5);
		actions[27][14]	= new Reduce(r5);
		actions[27][15]	= new Reduce(r5);
		actions[27][17]	= new Shift(13);
		actions[27][18]	= new Shift(10);
		actions[27][19]	= new Shift(18);
		actions[27][20]	= new Shift(11);

		actions[28][ 0]	= new Reduce(r9);
		actions[28][ 7]	= new Reduce(r9);
		actions[28][10]	= new Shift(16);
		actions[28][11]	= new Shift(17);
		actions[28][12]	= new Shift(19);
		actions[28][13]	= new Shift(21);
		actions[28][14]	= new Reduce(r9);
		actions[28][15]	= new Reduce(r9);
		actions[28][17]	= new Shift(13);
		actions[28][18]	= new Shift(10);
		actions[28][19]	= new Shift(18);
		actions[28][20]	= new Shift(11);

		actions[29][ 0]	= new Reduce(r3);
		actions[29][ 7]	= new Reduce(r3);
		actions[29][ 8]	= new Reduce(r3);
		actions[29][ 9]	= new Reduce(r3);
		actions[29][10]	= new Reduce(r3);
		actions[29][11]	= new Reduce(r3);
		actions[29][12]	= new Reduce(r3);
		actions[29][13]	= new Reduce(r3);
		actions[29][14]	= new Reduce(r3);
		actions[29][15]	= new Reduce(r3);
		actions[29][17]	= new Reduce(r3);
		actions[29][18]	= new Reduce(r3);
		actions[29][19]	= new Shift(18);
		actions[29][20]	= new Shift(11);

		actions[30][ 0]	= new Reduce(r10);
		actions[30][ 7]	= new Reduce(r10);
		actions[30][10]	= new Shift(16);
		actions[30][11]	= new Shift(17);
		actions[30][12]	= new Shift(19);
		actions[30][13]	= new Shift(21);
		actions[30][14]	= new Reduce(r10);
		actions[30][15]	= new Reduce(r10);
		actions[30][17]	= new Shift(13);
		actions[30][18]	= new Shift(10);
		actions[30][19]	= new Shift(18);
		actions[30][20]	= new Shift(11);

		actions[31][ 0]	= new Reduce(r2);
		actions[31][ 7]	= new Reduce(r2);
		actions[31][ 8]	= new Reduce(r2);
		actions[31][ 9]	= new Reduce(r2);
		actions[31][10]	= new Reduce(r2);
		actions[31][11]	= new Reduce(r2);
		actions[31][12]	= new Reduce(r2);
		actions[31][13]	= new Reduce(r2);
		actions[31][14]	= new Reduce(r2);
		actions[31][15]	= new Reduce(r2);
		actions[31][17]	= new Reduce(r2);
		actions[31][18]	= new Reduce(r2);
		actions[31][19]	= new Reduce(r2);
		actions[31][20]	= new Reduce(r2);

		actions[32][ 0]	= new Reduce(r4);
		actions[32][ 7]	= new Reduce(r4);
		actions[32][ 8]	= new Reduce(r4);
		actions[32][ 9]	= new Reduce(r4);
		actions[32][10]	= new Reduce(r4);
		actions[32][11]	= new Reduce(r4);
		actions[32][12]	= new Reduce(r4);
		actions[32][13]	= new Reduce(r4);
		actions[32][14]	= new Reduce(r4);
		actions[32][15]	= new Reduce(r4);
		actions[32][17]	= new Reduce(r4);
		actions[32][18]	= new Reduce(r4);
		actions[32][19]	= new Shift(18);
		actions[32][20]	= new Shift(11);

		actions[33][ 0]	= new Reduce(r12);
		actions[33][ 7]	= new Reduce(r12);
		actions[33][ 8]	= new Shift(14);
		actions[33][ 9]	= new Shift(12);
		actions[33][10]	= new Shift(16);
		actions[33][11]	= new Shift(17);
		actions[33][12]	= new Shift(19);
		actions[33][13]	= new Shift(21);
		actions[33][14]	= new Reduce(r12);
		actions[33][15]	= new Reduce(r12);
		actions[33][17]	= new Shift(13);
		actions[33][18]	= new Shift(10);
		actions[33][19]	= new Shift(18);
		actions[33][20]	= new Shift(11);

		actions[34][ 7]	= new Shift(35);
		actions[34][ 8]	= new Shift(14);
		actions[34][ 9]	= new Shift(12);
		actions[34][10]	= new Shift(16);
		actions[34][11]	= new Shift(17);
		actions[34][12]	= new Shift(19);
		actions[34][13]	= new Shift(21);
		actions[34][14]	= new Shift( 9);
		actions[34][15]	= new Shift(20);
		actions[34][17]	= new Shift(13);
		actions[34][18]	= new Shift(10);
		actions[34][19]	= new Shift(18);
		actions[34][20]	= new Shift(11);

		actions[35][ 0]	= new Reduce(r14);
		actions[35][ 7]	= new Reduce(r14);
		actions[35][ 8]	= new Reduce(r14);
		actions[35][ 9]	= new Reduce(r14);
		actions[35][10]	= new Reduce(r14);
		actions[35][11]	= new Reduce(r14);
		actions[35][12]	= new Reduce(r14);
		actions[35][13]	= new Reduce(r14);
		actions[35][14]	= new Reduce(r14);
		actions[35][15]	= new Reduce(r14);
		actions[35][17]	= new Reduce(r14);
		actions[35][18]	= new Reduce(r14);
		actions[35][19]	= new Reduce(r14);
		actions[35][20]	= new Reduce(r14);

		actions[36][ 0]	= new Reduce(r11);
		actions[36][ 7]	= new Reduce(r11);
		actions[36][ 8]	= new Shift(14);
		actions[36][ 9]	= new Shift(12);
		actions[36][10]	= new Shift(16);
		actions[36][11]	= new Shift(17);
		actions[36][12]	= new Shift(19);
		actions[36][13]	= new Shift(21);
		actions[36][14]	= new Reduce(r11);
		actions[36][15]	= new Reduce(r11);
		actions[36][17]	= new Shift(13);
		actions[36][18]	= new Shift(10);
		actions[36][19]	= new Shift(18);
		actions[36][20]	= new Shift(11);

    for (int i=0; i<NUM_STATES; i++)
      for (int j=0; j<NUM_TERMINALS; j++)
        if (actions[i][j] == null)
          actions[i][j] = new Error();
  }

}

