package roboforge.particles;

import java.util.*;
import roboforge.animation.*;



public abstract class BParticleSystemPooled extends BParticleSystemAbstract
{

  // Static variables...
  // -------------------------------------------------------------

  protected static final BParticleSystemException CREATE_EXCEPTION
    = new BParticleSystemException("Can't create particle!");



  // Member variables...
  // -------------------------------------------------------------

  protected BParticle[] particles;
    // The particle pool array.  The system can use up to n
    // particles, where n is particles.length.

  protected boolean[] isAlive;
    // Status indicators, one per particle in the pool.



  protected int n;
    // Total number of particles available.

  protected int used;
    // Number of allocated particles.

  protected int alive;
    // Number of 'live' particles in system.





  // Constructor...
  // -------------------------------------------------------------

  public BParticleSystemPooled(int n)
  {
    super();

    particles = new BParticle[n];
    isAlive   = new boolean[n];

    for (int i=0; i<n; i++)
      { isAlive[i] = false; }

    this.n     = n;
    this.alive = 0;
    this.used  = 0;
  }


  // Methods...
  // -------------------------------------------------------------

  public boolean isAlive()
    { return (alive > 0); }

  public boolean isDead()
    { return !isAlive(); }


  protected void updateParticles(float t, float dt)
  {
    BParticle p;

    for (int i=0; i<used; i++)
      if (isAlive[i])
      {
        // Determine current particle.
        p = particles[i];

        // Update the particle.
        p.update(t, dt);

        // Dispose of particle, if dead...
        if (p.isDead())
          killParticle(i);
      }
  }

  protected void killParticle(int i)
  {
    particles[i].stop();
    isAlive[i] = false;
    alive--;
  }


  protected void stopParticles()
  {
    // Kill any live particles.
    for (int i=0; i<used; i++)
      if (isAlive[i])
      {
        particles[i].stop();
        isAlive[i] = false;
      }

    alive = 0;
  }


  public BParticle emitParticle()
    throws BParticleSystemException
  {
    BParticle p = null;

    if (alive == used)
    {
      if (used == n)
        throw CREATE_EXCEPTION;

      p = createParticle();

      particles[used] = p;
      isAlive[used++] = true;
    }
    else
      for (int i=0; i<used; i++)
        if (!isAlive[i])
        {
          p = particles[i];
          isAlive[i] = true;
          break;
        }

    alive++;
    p.start();

    return p;
  }


  public void setPoolSize(int size)
  {
    if (size < 0)
      throw new IllegalArgumentException(
        "Illegal particle pool size: "+ size);

    if (size == n)
      return;

    // Kill any active particles, just in case.
    stopParticles();

    // Copy over any instantiated particles...
    BParticle[] old = particles;
    particles = new BParticle[size];

    if (n < size)
      System.arraycopy(old, 0, particles, 0, n);
    else
      System.arraycopy(old, 0, particles, 0, size);

    // Resize the isAlive array...
    isAlive = new boolean[size];
    for (int i=0; i<size; i++)
      { isAlive[i] = false; }

    n = size;
    used = Math.min(n, used);
  }


  protected void preloadParticles()
  {
  	// Preload all particles in the system,
		// for smooth animation later..

    // NB: Ensure that you have added in all aspects
    // 		 to this system prior to calling this.

  	used = n;

    for (int i=0; i<used; i++)
    	try
      	{ particles[i] = createParticle(); }
      catch (Exception e)
      	{ throw new RuntimeException("preloadParticles(): "+ e); }
  }


  /*
  public void cleanse()
  {
    super.cleanse();
    Iterator ps;

    ps = livePool.iterator();
    while (ps.hasNext())
      ((BParticle) ps.next()).cleanse();

    ps = deadPool.iterator();
    while (ps.hasNext())
      ((BParticle) ps.next()).cleanse();
  }
  */


  public String toString()
  	{ return getName()+ "["+ n+ "/"+ used+ "]"; }

}
