package roboforge.ui.layers;

import roboforge.ui.events.*;

import java.awt.*;
import java.util.Vector;
import javax.swing.*;


public class BLayerManager
  implements BActivateable
{


  // Class variables...
  // -------------------------------------------------------------
  
  public static final LayerVisitor ACTIVATOR   = new LayerActivator();
  public static final LayerVisitor DEACTIVATOR = new LayerDeactivator();
  public static final LayerVisitor CLOSER      = new LayerCloser();
  
  
  // Member variables...
  // -------------------------------------------------------------

  protected boolean singleLayerVisible;
  
  protected BLayer top;
  protected Vector tops;
  protected Vector list;

  protected JComponent holder;

  protected BActivationSupport activationSupport
    = new BActivationSupport(this);


  // Constructors...
  // -------------------------------------------------------------

  public BLayerManager(JComponent holder)
  { this(holder, true); }
  
  public BLayerManager(JComponent holder, boolean singleLayerVisible)
  {
    this.singleLayerVisible = singleLayerVisible;
    this.holder = holder;

    this.tops = new Vector(2);
    this.list = new Vector(5);
    this.top  = null;
  }



  // Accessor methods...
  // -------------------------------------------------------------

  public void add(BLayer l)
  { 
    if (contains(l))
      removeLayer(l);
    
    addLayer(l); 
  }

  public void remove(BLayer l)
    { removeLayer(l); }

  public void clear()
    { closeAllLayers(); }

  public boolean contains(BLayer l)
    { return tops.contains(l) || list.contains(l); }

  public void removeTopmost()
    { removeLayer(getTopmost()); } 
    // if (top != null) removeLayer(top); }

  public BLayer getTopmost()
  {
    if (!tops.isEmpty())
      return (BLayer) tops.lastElement();
    else if (!list.isEmpty())
      return (BLayer) list.lastElement();
    else
      return null;
  }

  public void moveToFront(BLayer l)
  {
    if (singleLayerVisible)
    {
      if (l != top)
      {
        removeLayer(l);
        addLayer(l);
      }
    }
    else
    {
      removeLayer(l);
      addLayer(l);
    }
  }


  // Layering implementation methods...
  // -------------------------------------------------------------

  protected void addLayer(BLayer l)
  {
    if (l.isAlwaysOnTop())
      tops.add(l);
    else
      list.add(l);

    if (singleLayerVisible)
    {
      boolean isReplacingTop =
        l.isAlwaysOnTop() || tops.isEmpty();

      if (isReplacingTop)
      {
        if (top != null)
          hide(top);

        show(l);
        top = l;
      }
    }
    else
      show(l);
  }


  protected void removeLayer(BLayer l)
  {
    if (l.isAlwaysOnTop())
      tops.remove(l);
    else
      list.remove(l);

    if (singleLayerVisible)
    {
      if (l == top)
      {
        hide(top);

        if (!tops.isEmpty())
          top = (BLayer) tops.lastElement();
        else if (!list.isEmpty())
          top = (BLayer) list.lastElement();
        else
          top = null;

        if (top != null)
          show(top);
      }
    }
    else
      hide(l);
  }

  protected void show(BLayer l)
  {
    if (l != null)
    {
      l.addTo(holder);

      if (isActive())
        l.requestActivate();
    }

    // holder.revalidate();
  }

  protected void hide(BLayer l)
  {
    if (l != null)
    {
      l.requestDeactivate();
      l.removeFrom(holder);
    }

    // holder.revalidate();
    // holder.repaint();
  }
  

  // Activation methods...
  // -------------------------------------------------------------

  public boolean isActive()
    { return activationSupport.isActive(); }

  public boolean isClosed()
    { return activationSupport.isClosed(); }


  public void requestActivate()
    { activationSupport.requestActivate(); }

  public void requestDeactivate()
    { activationSupport.requestDeactivate(); }

  public void requestClose()
    { activationSupport.requestClose(); }


  public void onActivate()
  {
    if (singleLayerVisible)
    {
      if (top != null)
        top.requestActivate();
    }
    else
      visitLayers(ACTIVATOR);

    activationSupport.fireActivate();
  }

  public void onDeactivate()
  {
    if (singleLayerVisible)
    {
      if (top != null)
        top.requestDeactivate();
    }
    else
      visitLayers(DEACTIVATOR);

    activationSupport.fireDeactivate();
  }

  public void onClose()
  {
    closeAllLayers();
    activationSupport.fireClose();
  }


  public void addActivationListener(BActivationListener l)
    { activationSupport.addActivationListener(l); }

  public void removeActivationListener(BActivationListener l)
    { activationSupport.addActivationListener(l); }



  // Activation support methods...
  // -------------------------------------------------------------

  protected void closeAllLayers()
  {
    visitLayers(CLOSER);
    
    tops.clear();
    list.clear();
  }

  protected void visitLayers(LayerVisitor visitor)
  {
    BLayer l;
    int n = tops.size();

    for (int i=0; i<n; i++)
    {
      l = (BLayer) tops.get(i);
      visitor.visit(l); // l.requestClose();
    }

    n = list.size();

    for (int i=0; i<n; i++)
    {
      l = (BLayer) list.get(i);
      visitor.visit(l); // l.requestClose();
    }
  }
  
  
  protected static abstract class LayerVisitor
    { public abstract void visit(BLayer l); }

  protected static class LayerActivator extends LayerVisitor
    { public void visit(BLayer l) { l.requestActivate(); } }
  
  protected static class LayerDeactivator extends LayerVisitor
    { public void visit(BLayer l) { l.requestDeactivate(); } }

  protected static class LayerCloser extends LayerVisitor
    { public void visit(BLayer l) { l.requestClose(); } }
  
}